"""Landlab component that simulates potential evapotranspiration rate.

Potential Evapotranspiration Component calculates spatially distributed
potential evapotranspiration based on input radiation factor (spatial
distribution of incoming radiation) using chosen method such as constant 
or Priestly Taylor. Ref: ASCE-EWRI Task Committee Report Jan 2005.

.. codeauthor:: Sai Nudurupati and Erkan Istanbulluoglu

Examples
--------
>>> import numpy as np
>>> from landlab import RasterModelGrid
>>> from landlab.components.pet import PotentialEvapotranspiration

Create a grid on which to calculate incident shortwave radiation

>>> grid = RasterModelGrid((5, 4), spacing=(0.2, 0.2))

The grid will need some input data. To check the names of the fields
that provide the input to this component, use the *input_var_names*
class property.

>>> PotentialEvapotranspiration.input_var_names
('radiation__ratio_to_flat_surface',)

Check the units for the fields.

>>> PotentialEvapotranspiration.var_units('radiation__ratio_to_flat_surface')
'None'

Create the input fields.

>>> grid['cell']['radiation__ratio_to_flat_surface'] = np.array([
...       0.38488566, 0.38488566,
...       0.33309785, 0.33309785,
...       0.37381705, 0.37381705])

If you are not sure about one of the input or output variables, you can
get help for specific variables.

>>> PotentialEvapotranspiration.var_help('radiation__ratio_to_flat_surface')
name: radiation__ratio_to_flat_surface
description:
  ratio of total incident shortwave radiation on sloped surface
  to flat surface
units: None
at: cell
intent: in

Check the output variable names

>>> sorted(PotentialEvapotranspiration.output_var_names)
['radiation__incoming_shortwave',
 'radiation__net',
 'radiation__net_longwave',
 'radiation__net_shortwave',
 'surface__potential_evapotranspiration_rate']

Instantiate the 'PotentialEvapotranspiration' component to work on this grid,
and run it.

>>> PET = PotentialEvapotranspiration(grid, method='PriestlyTaylor')

Run the *update* method to update output variables with current time
>>> current_time = 0.5
>>> PET.update(current_time)

>>> PET.grid.at_cell['radiation__incoming_shortwave']
array([ 33.09968448,  33.09968448,  28.64599771,  28.64599771,
        32.14779789,  32.14779789])

>>> PET.grid.at_cell['radiation__net']
array([ 13.9764353 ,  13.9764353 ,  12.09585347,  12.09585347,
        13.57449849,  13.57449849])

>>> PET.grid.at_cell['radiation__net_shortwave']
array([ 13.23987379,  13.23987379,  11.45839908,  11.45839908,
        12.85911915,  12.85911915])

>>> PET.grid.at_cell['surface__potential_evapotranspiration_rate']
array([ 0.25488065,  0.25488065,  0.22058551,  0.22058551,  0.24755075,
        0.24755075])
"""
from landlab import Component
from ...utils.decorators import use_file_name_or_kwds
import numpy as np


_VALID_METHODS = set(['Constant', 'PriestlyTaylor', 'MeasuredRadiationPT',
                      'Cosine'])


def _assert_method_is_valid(method):
    if method not in _VALID_METHODS:
        raise ValueError('%s: Invalid method name' % method)


class PotentialEvapotranspiration(Component):

    """Calculate spatially distributed potential evapotranspiration.

    Potential Evapotranspiration Component calculates spatially distributed
    potential evapotranspiration based on input radiation factor (spatial
    distribution of incoming radiation) using chosen method such as constant 
    or Priestly Taylor. Ref: Xiaochi et. al. 2013 for 'Cosine' method and
    ASCE-EWRI Task Committee Report Jan 2005 for 'PriestlyTaylor' method.
    Note: Calling 'PriestlyTaylor' method would generate/overwrite shortwave &
    longwave radiation fields.

    Construction::

        PotentialEvapotranspiration(grid, method='Cosine',
            priestly_taylor_const=1.26, albedo=0.6,
            latent_heat_of_vaporization=28.34, psychometric_const=0.066,
            stefan_boltzmann_const=0.0000000567, solar_const=1366.67,
            latitude=34., elevation_of_measurement=300, adjustment_coeff=0.18,
            lt=0., nd=365., MeanTmaxF=12., delta_d=5., **kwds)
        
    Parameters
    ----------
    grid: RasterModelGrid
        A grid.
    method: {'Constant', 'PriestlyTaylor', 'MeasuredRadiationPT', 'Cosine'}, optional
        Priestly Taylor method will spit out radiation outputs too.
    priestly_taylor_constant: float, optional
        Alpha used in Priestly Taylor method.
    albedo: float, optional
        Albedo.
    latent_heat_of_vaporization: float, optional
        Latent heat of vaporization for water Pwhv (Wd/(m*mm^2)).
    psychometric_const: float, optional
        Psychometric constant (kPa (deg C)^-1).
    stefan_boltzmann_const: float, optional
        Stefan Boltzmann's constant (W/(m^2K^-4)).
    solar_const: float, optional
        Solar constant (W/m^2).
    latitude: float, optional
        Latitude (radians).
    elevation_of_measurement: float, optional
        Elevation at which measurement was taken (m).
    adjustment_coeff: float, optional
        adjustment coeff to predict Rs from air temperature (deg C)^-0.5.
    lt: float, optional
        lag between peak TmaxF and solar forcing (days).
    nd: float, optional
        Number of days in year (days).
    MeanTmaxF: float, optional
        Mean annual rate of TmaxF (mm/d).
    delta_d: float, optional
        Calibrated difference between max & min daily TmaxF (mm/d).

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.components.pet import PotentialEvapotranspiration

    >>> grid = RasterModelGrid((5, 4), spacing=(0.2, 0.2))
    >>> PET = PotentialEvapotranspiration(grid)
    >>> PET.name
    'Potential Evapotranspiration'
    >>> PET.input_var_names
    ('radiation__ratio_to_flat_surface',)
    >>> sorted(PET.output_var_names)
    ['radiation__incoming_shortwave',
     'radiation__net',
     'radiation__net_longwave',
     'radiation__net_shortwave',
     'surface__potential_evapotranspiration_rate']
    >>> sorted(PET.units) # doctest: +NORMALIZE_WHITESPACE
    [('radiation__incoming_shortwave', 'W/m^2'),
     ('radiation__net', 'W/m^2'),
     ('radiation__net_longwave', 'W/m^2'),
     ('radiation__net_shortwave', 'W/m^2'),
     ('radiation__ratio_to_flat_surface', 'None'),
     ('surface__potential_evapotranspiration_rate', 'mm')]
    >>> PET.grid.number_of_cell_rows
    3
    >>> PET.grid.number_of_cell_columns
    2
    >>> PET.grid is grid
    True
    >>> pet_rate = grid.at_cell['surface__potential_evapotranspiration_rate']
    >>> np.allclose(pet_rate, 0.)
    True
    >>> grid['cell']['radiation__ratio_to_flat_surface'] = np.array([
    ...       0.38488566, 0.38488566,
    ...       0.33309785, 0.33309785,
    ...       0.37381705, 0.37381705])
    >>> current_time = 0.5
    >>> PET.update(current_time)
    >>> np.allclose(pet_rate, 0.)
    False
    """

    _name = 'Potential Evapotranspiration'

    _input_var_names = (
        'radiation__ratio_to_flat_surface',
    )

    _output_var_names = (
        'surface__potential_evapotranspiration_rate',
        'radiation__incoming_shortwave',
        'radiation__net_shortwave',
        'radiation__net_longwave',
        'radiation__net',
    )

    _var_units = {
        'radiation__ratio_to_flat_surface': 'None',
        'surface__potential_evapotranspiration_rate': 'mm',
        'radiation__incoming_shortwave': 'W/m^2',
        'radiation__net_shortwave': 'W/m^2',
        'radiation__net_longwave': 'W/m^2',
        'radiation__net': 'W/m^2',
    }

    _var_mapping = {
        'radiation__ratio_to_flat_surface': 'cell',
        'surface__potential_evapotranspiration_rate': 'cell',
        'radiation__incoming_shortwave': 'cell',
        'radiation__net_shortwave': 'cell',
        'radiation__net_longwave': 'cell',
        'radiation__net': 'cell',
    }

    _var_doc = {
        'radiation__ratio_to_flat_surface':
            'ratio of total incident shortwave radiation on sloped surface \
             to flat surface',
        'radiation__incoming_shortwave':
            'total incident shortwave radiation over the time step',
        'radiation__net_shortwave':
            'net incident shortwave radiation over the time step',
        'radiation__net_longwave':
            'net incident longwave radiation over the time step',
        'radiation__net':
            'net total radiation over the time step',
    }

    @use_file_name_or_kwds
    def __init__(self, grid, method='Cosine', priestly_taylor_const=1.26,
                 albedo=0.6, latent_heat_of_vaporization=28.34,
                 psychometric_const=0.066, stefan_boltzmann_const=0.0000000567,
                 solar_const=1366.67, latitude=34.,
                 elevation_of_measurement=300, adjustment_coeff=0.18,
                 lt=0., nd=365., MeanTmaxF=12., delta_d=5., **kwds):
        """
        Parameters
        ----------
        grid: RasterModelGrid
            A grid.
        method: {'Constant', 'PriestlyTaylor', 'MeasuredRadiationPT',
                 'Cosine'}, optional
            Priestly Taylor method will spit out radiation outputs too.
        priestly_taylor_constant: float, optional
            Alpha used in Priestly Taylor method.
        albedo: float, optional
            Albedo.
        latent_heat_of_vaporization: float, optional
            Latent heat of vaporization for water Pwhv (Wd/(m*mm^2)).
        psychometric_const: float, optional
            Psychometric constant (kPa (deg C)^-1).
        stefan_boltzmann_const: float, optional
            Stefan Boltzmann's constant (W/(m^2K^-4)).
        solar_const: float, optional
            Solar constant (W/m^2).
        latitude: float, optional
            Latitude (radians).
        elevation_of_measurement: float, optional
            Elevation at which measurement was taken (m).
        adjustment_coeff: float, optional
            adjustment coeff to predict Rs from air temperature (deg C)^-0.5.
        lt: float, optional
            lag between peak TmaxF and solar forcing (days).
        nd: float, optional
            Number of days in year (days).
        MeanTmaxF: float, optional
            Mean annual rate of TmaxF (mm/d).
        delta_d: float, optional
            Calibrated difference between max & min daily TmaxF (mm/d).
        """

        self._method = method
        # For Priestly Taylor
        self._alpha = priestly_taylor_const
        self._a = albedo
        self._pwhv = latent_heat_of_vaporization
        self._y = psychometric_const
        self._sigma = stefan_boltzmann_const
        self._Gsc = solar_const
        self._phi = (np.pi / 180.) * latitude
        self._z = elevation_of_measurement
        self._Krs = adjustment_coeff
        self._LT = lt
        self._ND = nd
        self._TmaxF_mean = MeanTmaxF
        self._DeltaD = delta_d

        _assert_method_is_valid(self._method)

        super(PotentialEvapotranspiration, self).__init__(grid, **kwds)

        for name in self._input_var_names:
            if name not in self.grid.at_cell:
                self.grid.add_zeros('cell', name, units=self._var_units[name])

        for name in self._output_var_names:
            if name not in self.grid.at_cell:
                self.grid.add_zeros('cell', name, units=self._var_units[name])

        self._cell_values = self.grid['cell']

    def update(self, current_time, const_potential_evapotranspiration=12.,
               Tmin=0., Tmax=1., Tavg=0.5, obs_radiation=350., **kwds):
        """Update fields with current conditions."""

        if self._method == 'Constant':
            self._PET_value = const_potential_evapotranspiration
        elif self._method == 'PriestlyTaylor':
            Tmin = Tmin
            Tmax = Tmax
            Tavg = Tavg
            self._PET_value = (
                self._PriestlyTaylor(current_time, Tmax, Tmin, Tavg))
            self._cell_values['radiation__incoming_shortwave'] = (
                self._Rs *
                self._cell_values['radiation__ratio_to_flat_surface'])
            self._cell_values['radiation__net_shortwave'] = (
                self._Rns *
                self._cell_values['radiation__ratio_to_flat_surface'])
            self._cell_values['radiation__net_longwave'] = (
                self._Rnl *
                self._cell_values['radiation__ratio_to_flat_surface'])
            self._cell_values['radiation__net'] = (
                self._Rn *
                self._cell_values['radiation__ratio_to_flat_surface'])
        elif self._method == 'MeasuredRadiationPT':
            Tavg = Tavg
            Robs = obs_radiation
            self._PET_value = self._MeasuredRadPT(Tavg, (1-self._a)*Robs)
        elif self._method == 'Cosine':
            self._J = np.floor((current_time - np.floor(current_time)) * 365.)
            self._PET_value = (
                max((self._TmaxF_mean + self._DeltaD / 2. *
                     np.cos((2 * np.pi) * (self._J - self._LT - self._ND / 2) /
                            self._ND)), 0.0))

        self._PET = (
            self._PET_value *
            self._cell_values['radiation__ratio_to_flat_surface'])
        self._cell_values['surface__potential_evapotranspiration_rate'][:] = (
            self._PET)

    def _PriestlyTaylor(self, current_time, Tmax, Tmin, Tavg):

        # Julian Day - ASCE-EWRI Task Committee Report, Jan-2005 - Eqn 25, (52)
        self._J = np.floor((current_time - np.floor(current_time)) * 365)

        # Saturation Vapor Pressure - ASCE-EWRI Task Committee Report,
        # Jan-2005 - Eqn 6, (37)
        self._es = 0.6108 * np.exp((17.27 * Tavg)/(237.7 + Tavg))

        # Actual Vapor Pressure - ASCE-EWRI Task Committee Report,
        # Jan-2005 - Eqn 8, (38)
        self._ea = 0.6108 * np.exp((17.27 * Tmin)/(237.7 + Tmin))

        # Slope of Saturation Vapor Pressure - ASCE-EWRI Task Committee Report,
        # Jan-2005 - Eqn 5, (36)
        self._delta = (4098.0 * self._es)/((237.3 + Tavg) ** 2.0)

        # Solar Declination Angle - ASCE-EWRI Task Committee Report,
        # Jan-2005 - Eqn 24,(51)
        self._sdecl = 0.409 * np.sin(((np.pi / 180.0) * self._J) - 1.39)

        # Inverse Relative Distance Factor - ASCE-EWRI Task Committee Report,
        # Jan-2005 - Eqn 23,(50)
        self._dr = 1 + (0.033 * np.cos(np.pi / 180.0 * self._J))

        # To calculate ws - ASCE-EWRI Task Committee Report,
        # Jan-2005 - Eqn 29,(61)
        self._x = 1.0 - (((np.tan(self._phi)) ** 2.0) *
                         (np.tan(self._sdecl) ** 2.0))
        if self._x <= 0:
            self._x = 0.00001
            # Sunset Hour Angle - ASCE-EWRI Task Committee Report,
            # Jan-2005 - Eqn 28,(60)
        self._ws = ((np.pi / 2.0) -
                    np.arctan((- 1 * np.tan(self._phi) *
                               np.tan(self._sdecl)) / (self._x ** 2.0)))

        # Extraterrestrial radmodel.docx - ASCE-EWRI Task Committee Report,
        # Jan-2005 - Eqn 21, (48)
        self._Ra = (
                    11.57 * (24.0 / np.pi) * 4.92 * self._dr *
                    ((self._ws * np.sin(self._phi) * np.sin(self._sdecl)) +
                     (np.cos(self._phi) * np.cos(self._sdecl) *
                     (np.sin(self._ws)))))

        # Clear-sky Solar Radiation - ASCE-EWRI Task Committee Report,
        # Jan-2005 - Eqn 19, (47)
        self._Rso = (0.75 + ((2.0 * (10 ** (- 5.0))) * self._z)) * self._Ra
        self._Rs = min(self._Krs * self._Ra * np.sqrt(Tmax - Tmin), self._Rso)

        # Net Short Wave Radiation - ASCE-EWRI Task Committee Report,
        # Jan-2005 - Eqn 16, (43)
        self._Rns = self._Rs * (1 - self._a)

        # Relative Cloudiness - ASCE-EWRI Task Committee Report,
        # Jan-2005 - Page 20,35
        if self._Rso > 0:
            self._u = self._Rs / self._Rso
        else:
            self._u = 0

        if self._u < 0.3:
            self._u = 0.3
        elif self._u > 1:
            self._u = 1.0

        # Cloudiness Function - ASCE-EWRI Task Committee Report,
        # Jan-2005 - Eqn 18, (45)
        self._fcd = (1.35 * self._u) - 0.35

        # Net Long Wave Radiation - ASCE-EWRI Task Committee Report,
        # Jan-2005 - Eqn 17, (44)
        self._Rnl = (self._sigma *
                     self._fcd * (0.34 - (0.14 * np.sqrt(self._ea)) *
                                  (((Tmax + 273.16) ** 4.0 +
                                    (Tmin + 273.16) ** 4.0) / 2.0)))

        # Net Radiation - ASCE-EWRI Task Committee Report,
        # Jan-2005 - Eqn 15, (42)
        self._Rn = self._Rns - self._Rnl

        self._ETp = max(self._alpha * (self._delta / (self._delta + self._y)) *
                        (self._Rn / self._pwhv), 0)

        return self._ETp

    def _MeasuredRadPT(self, Tavg, Rnobs):
        # Saturation Vapor Pressure - ASCE-EWRI Task Committee Report,
        # Jan-2005 - Eqn 6, (37)
        self._es = 0.6108 * np.exp((17.27 * Tavg)/(237.7 + Tavg))

        # Slope of Saturation Vapor Pressure - ASCE-EWRI Task Committee Report,
        # Jan-2005 - Eqn 5, (36)
        self._delta = (4098.0 * self._es) / ((237.3 + Tavg) ** 2.0)
        self._ETp = max(self._alpha * (self._delta / (self._delta + self._y)) *
                        (Rnobs / self._pwhv), 0)

        return self._ETp



# -*- coding: utf-8 -*-
"""
test_sink_filler:

Created on Tues Oct 20, 2015

@author: dejh
"""

import landlab
from landlab import RasterModelGrid, FieldError
from landlab.components.flow_routing import FlowRouter
from landlab.components.sink_fill import SinkFiller
from numpy import sin, pi
import numpy as np  # for use of np.round
from numpy.testing import assert_array_equal, assert_array_almost_equal
from landlab import BAD_INDEX_VALUE as XX
from nose.tools import (with_setup, assert_true, assert_false, assert_raises,
                        assert_almost_equal, assert_equal)
try:
    from nose.tools import (assert_is, assert_set_equal, assert_dict_equal)
except ImportError:
    from landlab.testing.tools import (assert_is, assert_set_equal,
                                       assert_dict_equal)


def setup_dans_grid1():
    """
    Create a 7x7 test grid with a well defined hole in it.
    """
    global hf, mg
    global z, r_new, r_old, A_new, A_old, s_new, depr_outlet_target
    global lake, outlet, lake_code, outlet_array

    mg = RasterModelGrid(7, 7, 1.)

    z = np.array([0.0,  0.0,  0.0,  0.0,  0.0,  0.0,  0.0,
                  0.0,  2.0,  2.0,  2.0,  2.0,  2.0,  0.0,
                  0.0,  2.0,  1.6,  1.5,  1.6,  2.0,  0.0,
                  0.0,  2.0,  1.7,  1.6,  1.7,  2.0,  0.0,
                  0.0,  2.0,  1.8,  2.0,  2.0,  2.0,  0.0,
                  0.0,  1.0,  0.6,  1.0,  1.0,  1.0,  0.0,
                  0.0,  0.0, -0.5,  0.0,  0.0,  0.0,  0.0]).flatten()

    depr_outlet_target = np.array([XX, XX, XX, XX, XX, XX, XX,
                                   XX, XX, XX, XX, XX, XX, XX,
                                   XX, XX, 30, 30, 30, XX, XX,
                                   XX, XX, 30, 30, 30, XX, XX,
                                   XX, XX, XX, XX, XX, XX, XX,
                                   XX, XX, XX, XX, XX, XX, XX,
                                   XX, XX, XX, XX, XX, XX, XX]).flatten()

    lake = np.array([16, 17, 18, 23, 24, 25])
    outlet = 30
    lake_code = 17
    outlet_array = np.array([outlet])

    mg.add_field('node', 'topographic__elevation', z, units='-')

    hf = SinkFiller(mg)


def setup_dans_grid2():
    """
    Create a 10x10 test grid with a well defined hole in it, from a flat
    surface.
    """
    global hf, mg
    global z, depr_outlet_target
    global lake, outlet, lake_code, outlet_array

    lake = np.array([44, 45, 46, 54, 55, 56, 64, 65, 66])
    outlet = 35  # shouldn't be needed
    lake_code = 44
    outlet_array = np.array([outlet])

    mg = RasterModelGrid(10, 10, 1.)

    z = np.ones(100, dtype=float)
    z[lake] = 0.

    depr_outlet_target = np.empty(100, dtype=float)
    depr_outlet_target.fill(XX)
    depr_outlet_target = XX  # not well defined in this simplest case...?

    mg.add_field('node', 'topographic__elevation', z, units='-')

    hf = SinkFiller(mg)


def setup_dans_grid3():
    """
    Create a 10x10 test grid with two well defined holes in it, into an
    inclined surface.
    """
    global hf, fr, mg
    global z, depr_outlet_target
    global lake, lake1, lake2, outlet, outlet_array, rcvr

    lake1 = np.array([34, 35, 36, 44, 45, 46, 54, 55, 56])
    lake2 = np.array([77, 78, 87, 88])
    guard_nodes = np.array([23, 33, 53, 63])
    lake = np.concatenate((lake1, lake2))
    outlet = 35  # shouldn't be needed
    outlet_array = np.array([outlet])

    mg = RasterModelGrid(10, 10, 1.)

    z = np.ones(100, dtype=float)
    # add slope
    z += mg.node_x
    z[guard_nodes] += 0.001
    z[lake] = 0.

    depr_outlet_target = np.empty(100, dtype=float)
    depr_outlet_target.fill(XX)
    depr_outlet_target = XX  # not well defined in this simplest case...?

    mg.add_field('node', 'topographic__elevation', z, units='-')

    fr = FlowRouter(mg)


def setup_dans_grid4():
    """
    Create a 10x10 test grid with two well defined holes in it, into an
    inclined surface. This time, one of the holes is a stupid shape, which
    will require the component to arrange flow back "uphill".
    """
    global hf, fr, mg
    global z, depr_outlet_target
    global lake, lake1, lake2, outlet, outlet_array

    lake1 = np.array([34, 35, 36, 44, 45, 46, 54, 55, 56, 65, 74])
    lake2 = np.array([78, 87, 88])
    guard_nodes = np.array([23, 33, 53, 63, 73, 83])
    lake = np.concatenate((lake1, lake2))
    outlet = 35  # shouldn't be needed
    outlet_array = np.array([outlet])

    mg = RasterModelGrid(10, 10, 1.)

    z = np.ones(100, dtype=float)
    # add slope
    z += mg.node_x
    z[guard_nodes] += 0.001  # forces the flow out of a particular node
    z[lake] = 0.

    depr_outlet_target = np.empty(100, dtype=float)
    depr_outlet_target.fill(XX)
    depr_outlet_target = XX  # not well defined in this simplest case...?

    mg.add_field('node', 'topographic__elevation', z, units='-')

    fr = FlowRouter(mg)


def setup_dans_grid5():
    """
    Create a 10x10 test grid with two well defined holes in it, into an
    inclined surface. This time, one of the holes is a stupid shape, which
    will require the component to arrange flow back "uphill". Exactly as
    V4, but this version tests D4 routing.

    Notes
    -----
    Here is the elevation grid:

    1.      2.      3.      4.      5.      6.      7.      8.      9.     10.
    1.      2.      3.      4.      5.      6.      7.      8.      9.     10.
    1.      2.      3.      4.001   5.      6.      7.      8.      9.     10.
    1.      2.      3.      4.001   0.      0.      0.      8.      9.     10.
    1.      2.      3.      4.      5.      6.      7.      8.      9.     10.
    1.      2.      3.      4.001   0.      0.      0.      8.      9.     10.
    1.      2.      3.      4.001   5.      0.      7.      8.      9.     10.
    1.      2.      3.      4.001   0.      6.      7.      8.      0.     10.
    1.      2.      3.      4.001   5.      6.      7.      0.      0.     10.
    1.      2.      3.      4.      5.      6.      7.      8.      9.     10.
    """
    global hf, fr, mg
    global z, depr_outlet_target
    global lake, lake1, lake2, outlet, outlet_array

    lake1 = np.array([34, 35, 36, 44, 45, 46, 54, 55, 56, 65, 74])
    lake2 = np.array([78, 87, 88])
    guard_nodes = np.array([23, 33, 53, 63, 73, 83])
    lake = np.concatenate((lake1, lake2))
    outlet = 35  # shouldn't be needed
    outlet_array = np.array([outlet])

    mg = RasterModelGrid(10, 10, 1.)

    z = np.ones(100, dtype=float)
    # add slope
    z += mg.node_x
    z[guard_nodes] += 0.001  # forces the flow out of a particular node
    z[lake] = 0.

    depr_outlet_target = np.empty(100, dtype=float)
    depr_outlet_target.fill(XX)
    depr_outlet_target = XX  # not well defined in this simplest case...?

    mg.add_field('node', 'topographic__elevation', z, units='-')

    fr = FlowRouter(mg)


@with_setup(setup_dans_grid1)
def check_fields(grid):
    """
    Check to make sure the right fields have been created.
    """
    assert_array_equal(z, mg.at_node['topographic__elevation'])
    assert_array_equal(np.zeros(mg.number_of_nodes),
                       mg.at_node['sediment_fill__depth'])
    assert_raises(FieldError, mg.at_node['drainage_area'])


@with_setup(setup_dans_grid1)
def test_get_lake_ext_margin():
    lake = np.array([16, 17, 23, 24, 25, 30, 31, 32])
    ext_margin_returned = hf._get_lake_ext_margin(lake)
    ext_margin = np.array([8, 9, 10, 11, 15, 18, 19, 22, 26, 29, 33, 36, 37,
                           38, 39, 40])
    assert_array_equal(ext_margin_returned, ext_margin)


@with_setup(setup_dans_grid1)
def test_get_lake_int_margin():
    lake = np.array([16, 17, 18, 23, 24, 25, 26, 30, 31, 32])
    ext_margin = np.array([8, 9, 10, 11, 12, 15, 19, 20, 22, 27, 29, 33, 34,
                           36, 37, 38, 39, 40])
    int_margin_returned = hf._get_lake_int_margin(lake, ext_margin)
    int_margin = np.array([16, 17, 18, 23, 25, 26, 30, 31, 32])
    assert_array_equal(int_margin_returned, int_margin)


@with_setup(setup_dans_grid1)
def test_drainage_directions_change():
    lake = np.array([22, 23])
    old_elevs = np.ones(49, dtype=float)
    old_elevs[lake] = 0.
    new_elevs = old_elevs.copy()
    new_elevs[40] = 2.
    cond = hf.drainage_directions_change(lake, old_elevs, new_elevs)
    assert_false(cond)
    new_elevs[23] = 0.5
    cond = hf.drainage_directions_change(lake, old_elevs, new_elevs)
    assert_false(cond)
    new_elevs[23] = 1.
    cond = hf.drainage_directions_change(lake, old_elevs, new_elevs)
    assert_false(cond)
    new_elevs[23] = 1.2
    cond = hf.drainage_directions_change(lake, old_elevs, new_elevs)
    assert_true(cond)


@with_setup(setup_dans_grid1)
def test_add_slopes():
    new_z = z.copy()
    outlet_elev = z[outlet]
    hf._elev[lake] = outlet_elev
    rt2 = np.sqrt(2.)
    slope_to_add = 0.1
    lake_map = np.empty_like(z)
    lake_map.fill(XX)
    lake_map[lake] = lake_code
    hf._lf._lake_map = lake_map
    hf.lake_nodes_treated = np.array([], dtype=int)
    dists = mg.calc_distances_of_nodes_to_point((mg.node_x[outlet],
                                                mg.node_y[outlet]))
    new_z[lake] = outlet_elev
    new_z[lake] += dists[lake]*slope_to_add
    # test the ones we can do easily analytically separately
    straight_north = np.array([23, 16])
    off_angle = 24
    elevs_out, lake_out = hf._add_slopes(slope_to_add, outlet, lake_code)
    assert_array_equal(slope_to_add*(np.arange(2.)+1.)+outlet_elev,
                       elevs_out[straight_north])
    assert_almost_equal(slope_to_add*rt2+outlet_elev, elevs_out[off_angle])
    assert_array_equal(new_z, elevs_out)
    assert_array_equal(lake, lake_out)


@with_setup(setup_dans_grid2)
def test_filler_flat():
    """
    Very simple, though possibly degerate, case, filling a 3x3 hole up to
    the flat surface surrounding it.
    """
    hf.fill_pits()
    assert_array_equal(hf._elev[lake], np.ones(9, dtype=float))
    assert_array_equal(mg.at_node['topographic__elevation'][lake],
                       np.ones(9, dtype=float))


@with_setup(setup_dans_grid3)
def test_filler_inclined():
    """
    Tests a flat fill into an inclined surface, with two holes.
    """
    hf = SinkFiller(mg)
    hf.fill_pits()
    assert_array_equal(mg.at_node['topographic__elevation'][lake1],
                       np.ones(9, dtype=float)*4.)
    assert_array_equal(mg.at_node['topographic__elevation'][lake2],
                       np.ones(4, dtype=float)*7.)


@with_setup(setup_dans_grid3)
def test_filler_inclined2():
    """
    Tests an inclined fill into an inclined surface, with two holes.
    """
    z_init = z.copy()
    hf = hf = SinkFiller(mg, apply_slope=True)
    hf.fill_pits()
    hole1 = np.array([4.00009091, 4.00018182, 4.00027273, 4.00063636,
                      4.00045455, 4.00036364, 4.00081818, 4.00072727,
                      4.00054545])
    hole2 = np.array([7.16666667, 7.33333333, 7.66666667, 7.5])
    assert_array_almost_equal(mg.at_node['topographic__elevation'][lake1],
                              hole1)
    assert_array_almost_equal(mg.at_node['topographic__elevation'][lake2],
                              hole2)
    fr.route_flow()
    assert_equal(mg.at_node['flow__sink_flag'][mg.core_nodes].sum(), 0)


@with_setup(setup_dans_grid4)
def test_stupid_shaped_hole():
    """
    Tests inclined fill into a surface with a deliberately awkward shape.
    """
    hf = SinkFiller(mg, apply_slope=True)
    hf.fill_pits()
    hole1 = np.array([4.00007692, 4.00015385, 4.00023077, 4.00053846,
                      4.00038462, 4.00030769, 4.00069231, 4.00061538,
                      4.00046154, 4.00076923, 4.00084615])
    hole2 = np.array([7.4, 7.2, 7.6])
    # print this to check out the funky drainage arrangement...
    # print(mg.at_node['topographic__elevation'].reshape((10, 10))[3:8, 4:7])
    assert_array_almost_equal(mg.at_node['topographic__elevation'][lake1],
                              hole1)
    assert_array_almost_equal(mg.at_node['topographic__elevation'][lake2],
                              hole2)
    fr.route_flow()
    assert_equal(mg.at_node['flow__sink_flag'][mg.core_nodes].sum(), 0)


@with_setup(setup_dans_grid5)
def test_D4_routing():
    """
    Tests inclined fill into a surface with a deliberately awkward shape.
    This is testing D4 routing.
    """
    hf = SinkFiller(mg, routing='D4', apply_slope=True)
    hf.fill_pits()
    hole1 = np.array([4.00016667, 4.00025, 4.00033333, 4.00008333, 4.00041667,
                      4.0005, 4.00083333, 4.00066667, 4.00058333, 4.00075,
                      4.334])
    hole2 = np.array([7.6, 7.2, 7.4])
    # np.array([34, 35, 36, 44, 45, 46, 54, 55, 56, 65, 74])
    # print this to check out the funky drainage arrangement...
    # print(mg.at_node['topographic__elevation'].reshape((10, 10))[3:8, 4:7])
    assert_array_almost_equal(mg.at_node['topographic__elevation'][lake1],
                              hole1)
    assert_array_almost_equal(mg.at_node['topographic__elevation'][lake2],
                              hole2)
    fr.route_flow(method='D4')
    assert_equal(mg.at_node['flow__sink_flag'][mg.core_nodes].sum(), 0)


@with_setup(setup_dans_grid5)
def test_D4_filling():
    """
    Tests inclined fill into a surface with a deliberately awkward shape.
    This is testing D4 without inclining the surface.
    """
    hf = SinkFiller(mg, routing='D4')
    hf.fill_pits()
    hole1 = 4.*np.ones_like(lake1, dtype=float)
    hole1[-1] += 0.001
    hole2 = 7.*np.ones_like(lake2, dtype=float)
    # np.array([34, 35, 36, 44, 45, 46, 54, 55, 56, 65, 74])
    # print this to check out the funky drainage arrangement...
    # print(mg.at_node['topographic__elevation'].reshape((10, 10))[3:8, 4:7])
    assert_array_almost_equal(mg.at_node['topographic__elevation'][lake1],
                              hole1)
    assert_array_almost_equal(mg.at_node['topographic__elevation'][lake2],
                              hole2)



"""
simple script to run speed tests of various functions in model grid
"""

from __future__ import print_function
import landlab as ll
import time


def main():
    mg = ll.RasterModelGrid(20, 30, 1.0)

    nt = 1000

    s = mg.zeros(at='node')
    g = mg.zeros(at='link')
    divg = mg.zeros(at='node')

    start_time = time.time()

    for i in range(nt):

        g = mg.calc_grad_at_link(s, out=g)

    time1 = time.time()

    time2 = time.time()

    for i in range(nt):

        divg = mg.calculate_flux_divergence_at_nodes(g, divg)

    time3 = time.time()

    for i in range(nt):

        divg = mg.calculate_flux_divergence_at_nodes_slow(g, divg)

    time4 = time.time()

    for i in range(nt):

        divg = mg.calculate_flux_divergence_at_active_cells(g)

    time5 = time.time()

    for i in range(nt):

        divg = mg.calculate_flux_divergence_at_active_cells_slow(g)

    time6 = time.time()

    print('Elapsed time with fast gradient algo: ' + str(time1 - start_time))
    print('Elapsed time with slow gradient algo: ' + str(time2 - time1))
    print('Elapsed time with fast node-divergence algo: ' + str(time3 - time2))
    print('Elapsed time with slow node-divergence algo: ' + str(time4 - time3))
    print('Elapsed time with fast activecell-divergence algo: ' + str(time5 - time4))
    print('Elapsed time with slow activecell-divergence algo: ' + str(time6 - time5))


if __name__ == '__main__':
    main()



import numpy as np
from numpy.testing import assert_array_equal
from nose.tools import (assert_equal, assert_raises, raises, assert_true,
                        assert_false)
from nose.tools import with_setup

try:
    from nose.tools import assert_is, assert_is_instance
except ImportError:
    from landlab.testing.tools import assert_is, assert_is_instance

from landlab import RasterModelGrid
from landlab import BAD_INDEX_VALUE as X


def setup_grid():
    """
    These tests use a grid that 4x5 nodes::

     15------16------17------18------19
      |       |       |       |       |
      |       |       |       |       |
      |       |       |       |       |
     10------11------12------13------14
      |       |       |       |       |
      |       |       |       |       |
      |       |       |       |       |
      5-------6-------7-------8-------9
      |       |       |       |       |
      |       |       |       |       |
      |       |       |       |       |
      0-------1-------2-------3-------4
    """
    from landlab import RasterModelGrid
    globals().update({
        'rmg': RasterModelGrid(4, 5, 1.)
    })


def test_init_with_kwds_classic():
    grid = RasterModelGrid(num_rows=4, num_cols=5, dx=1.)

    assert_equal(grid.number_of_node_rows, 4)
    assert_equal(grid.number_of_node_columns, 5)
    assert_equal(grid.dy, 1)
    assert_equal(grid.dx, 1)

    grid = RasterModelGrid(3, 7, 2)

    assert_equal(grid.number_of_node_rows, 3)
    assert_equal(grid.number_of_node_columns, 7)
    assert_equal(grid.dy, 2.)
    assert_equal(grid.dx, 2.)


def test_init_new_style():
    grid = RasterModelGrid((4, 5), spacing=2)

    assert_equal(grid.number_of_node_rows, 4)
    assert_equal(grid.number_of_node_columns, 5)
    assert_equal(grid.dy, 2.)
    assert_equal(grid.dx, 2.)

    grid = RasterModelGrid((4, 5))

    assert_equal(grid.number_of_node_rows, 4)
    assert_equal(grid.number_of_node_columns, 5)
    assert_equal(grid.dy, 1.)
    assert_equal(grid.dx, 1.)


def test_spacing_is_float():
    grid = RasterModelGrid((4, 5))
    assert_equal(grid.dy, 1.)
    assert_is_instance(grid.dy, float)
    assert_equal(grid.dx, 1.)
    assert_is_instance(grid.dx, float)

    grid = RasterModelGrid((4, 5), spacing=2)
    assert_equal(grid.dy, 2.)
    assert_is_instance(grid.dy, float)
    assert_equal(grid.dx, 2.)
    assert_is_instance(grid.dx, float)


@with_setup(setup_grid)
def test_grid_dimensions():
    """Test extent of grid with unit spacing."""
    assert_equal(rmg.extent[0], rmg.number_of_node_rows - 1)
    assert_equal(rmg.extent[1], rmg.number_of_node_columns - 1)


def test_grid_dimensions_non_unit_spacing():
    """Test extent of grid with non-unit spacing."""
    rmg = RasterModelGrid((4, 5), spacing=2.)
    assert_equal(rmg.extent[0], 6.)
    assert_equal(rmg.extent[1], 8.)


@with_setup(setup_grid)
def test_nodes_around_point():
    surrounding_ids = rmg.nodes_around_point(2.1, 1.1)
    assert_array_equal(surrounding_ids, np.array([7, 12, 13, 8]))

    surrounding_ids = rmg.nodes_around_point(2.1, .9)
    assert_array_equal(surrounding_ids, np.array([2, 7, 8, 3]))


@with_setup(setup_grid)
def test_neighbor_list_with_scalar_arg():
    assert_array_equal(rmg.active_neighbors_at_node(6), np.array([7, 11, 5, 1]))
    assert_array_equal(rmg.active_neighbors_at_node(-1), np.array([X, X, X, X]))
    assert_array_equal(rmg.active_neighbors_at_node(-2), np.array([X, X, X, 13]))


@with_setup(setup_grid)
def test_neighbor_list_with_array_arg():
    assert_array_equal(rmg.active_neighbors_at_node([6, -1]),
                       np.array([[7, 11, 5, 1], [X, X, X, X]]))


@with_setup(setup_grid)
def test_neighbor_list_with_no_args():
    expected = np.array([
        [X,  X,  X,  X], [X,  6,  X,  X], [X,  7,  X,  X], [X,  8,  X,  X],
        [X,  X,  X,  X],
        [6,  X,  X,  X], [7, 11,  5,  1], [8, 12,  6,  2], [9, 13,  7,  3],
        [X,  X,  8,  X],
        [11,  X,  X,  X], [12, 16, 10,  6], [13, 17, 11,  7], [14, 18, 12,  8],
        [X,  X, 13,  X],
        [X,  X,  X,  X], [X,  X,  X, 11], [X,  X,  X, 12], [X,  X,  X, 13],
        [X,  X,  X,  X]])

    assert_array_equal(rmg.active_neighbors_at_node(), expected)


@with_setup(setup_grid)
def test_node_x():
    assert_array_equal(rmg.node_x, np.array([0., 1., 2., 3., 4.,
                                             0., 1., 2., 3., 4.,
                                             0., 1., 2., 3., 4.,
                                             0., 1., 2., 3., 4.]))


@with_setup(setup_grid)
def test_node_y():
    assert_array_equal(rmg.node_y, np.array([0., 0., 0., 0., 0.,
                                             1., 1., 1., 1., 1.,
                                             2., 2., 2., 2., 2.,
                                             3., 3., 3., 3., 3.]))


@with_setup(setup_grid)
@raises(ValueError)
def test_node_x_is_immutable():
    rmg.node_x[0] = 0


@with_setup(setup_grid)
@raises(ValueError)
def test_node_y_is_immutable():
    rmg.node_y[0] = 0


@with_setup(setup_grid)
def test_node_axis_coordinates():
    assert_is(rmg.node_axis_coordinates(axis=0).base, rmg.node_y.base)
    assert_is(rmg.node_axis_coordinates(axis=1).base, rmg.node_x.base)
    assert_is(rmg.node_axis_coordinates(axis=-1).base, rmg.node_x.base)
    assert_is(rmg.node_axis_coordinates(axis=-2).base, rmg.node_y.base)


@with_setup(setup_grid)
def test_diagonal_list():
    assert_array_equal(rmg._get_diagonal_list(6), np.array([12, 10, 0, 2]))
    assert_array_equal(rmg._get_diagonal_list(-1), np.array([X, X, 13, X]))
    assert_array_equal(rmg._get_diagonal_list([6, -1]),
                       np.array([[12, 10, 0, 2], [X, X, 13, X]]))
    assert_array_equal(
        rmg._get_diagonal_list(),
        np.array([[6, X, X, X], [7, 5, X, X], [8, 6, X, X],
                  [9, 7, X, X], [X, 8, X, X],
                  [11, X, X, 1], [12, 10,  0,  2], [13, 11,  1,  3],
                  [14, 12,  2,  4], [X, 13, 3, X],
                  [16, X, X, 6], [17, 15,  5,  7], [18, 16,  6,  8],
                  [19, 17,  7,  9], [X, 18, 8, X],
                  [X, X, X, 11], [X, X, 10, 12], [X, X, 11, 13],
                  [X, X, 12, 14], [X, X, 13, X]]))


@with_setup(setup_grid)
def test_diagonal_list_boundary():
    assert_array_equal(rmg._get_diagonal_list(0), np.array([6, X, X, X]))


@with_setup(setup_grid)
def test_node_is_core():
    for cell_id in [0, 1, 2, 3, 4, 5, 9, 10, 14, 15, 16, 17, 18, 19]:
        assert_false(rmg.node_is_core(cell_id))
    for cell_id in [6, 7, 8, 11, 12, 13]:
        assert_true(rmg.node_is_core(cell_id))


@with_setup(setup_grid)
def test_get_interior_cells():
    assert_array_equal(rmg.node_at_core_cell,
                       np.array([6, 7, 8, 11, 12, 13]))


@with_setup(setup_grid)
def test_active_links():
    assert_equal(rmg.number_of_active_links, 17)
    assert_array_equal(rmg.active_links,
                       np.array([ 5,  6,  7,
                                  9, 10, 11, 12,
                                 14, 15, 16,
                                 18, 19, 20, 21,
                                 23, 24, 25]))


#@with_setup(setup_grid)
#def test_active_link_fromnode():
#    assert_array_equal(rmg._activelink_fromnode,
#                       np.array([1, 2, 3, 6, 7, 8, 11, 12, 13,
#                                 5, 6, 7, 8, 10, 11, 12, 13]))
#
#
#@with_setup(setup_grid)
#def test_active_link_tonode():
#    assert_array_equal(rmg._activelink_tonode,
#                       np.array([6, 7, 8, 11, 12, 13, 16, 17, 18,
#                                 6, 7, 8, 9, 11, 12, 13, 14]))


@with_setup(setup_grid)
def test_active_link_num_inlink():
    assert_array_equal(rmg._node_numactiveinlink,
                       np.array([0, 0, 0, 0, 0,
                                 0, 2, 2, 2, 1,
                                 0, 2, 2, 2, 1,
                                 0, 1, 1, 1, 0]))


@with_setup(setup_grid)
def test_active_link_num_outlink():
    assert_array_equal(rmg._node_numactiveoutlink, np.array([0, 1, 1, 1, 0,
                                                            1, 2, 2, 2, 0,
                                                            1, 2, 2, 2, 0,
                                                            0, 0, 0, 0, 0]))


@with_setup(setup_grid)
def test_active_inlink_matrix():
    assert_array_equal(rmg._node_active_inlink_matrix,
                       np.array([[-1, -1, -1, -1, -1,
                                  -1,  0,  1,  2, -1,
                                  -1,  3,  4,  5, -1,
                                  -1,  6, 7,  8, -1],
                                 [-1, -1, -1, -1, -1,
                                  -1,  9, 10, 11, 12,
                                  -1, 13, 14, 15, 16,
                                  -1, -1, -1, -1, -1]]))


@with_setup(setup_grid)
def test_active_outlink_matrix():
    assert_array_equal(
        rmg._node_active_outlink_matrix,
        np.array([[-1,  0,  1,  2, -1,
                   -1,  3,  4,  5, -1,
                   -1,  6,  7,  8, -1,
                   -1, -1, -1, -1, -1],
                  [-1, -1, -1, -1, -1,
                   9, 10, 11, 12, -1,
                   13, 14, 15, 16, -1,
                   -1, -1, -1, -1, -1]]))


@with_setup(setup_grid)
def test__active_links_at_node_scalar_interior():
    assert_array_equal(rmg._active_links_at_node([6]),
                       np.array([[5, 9, 14, 10]]).T)


@with_setup(setup_grid)
def test__active_links_at_node_scalar_boundary():
    assert_array_equal(rmg._active_links_at_node([1]),
                       np.array([[-1, -1,  5, -1]]).T)


@with_setup(setup_grid)
def test_active_node_with_array_arg():
    assert_array_equal(rmg._active_links_at_node([6, 7]),
                       np.array([[5,  9, 14, 10],
                                 [6, 10, 15, 11]]).T)


@with_setup(setup_grid)
def test__active_links_at_node_with_no_args():
    assert_array_equal(
        rmg._active_links_at_node(),
        np.array([[-1, -1, -1, -1, -1, -1,  5,  6,  7, -1,
                   -1, 14, 15, 16, -1, -1, 23, 24, 25, -1],
                  [-1, -1, -1, -1, -1, -1,  9, 10, 11, 12,
                   -1, 18, 19, 20, 21, -1, -1, -1, -1, -1],
                  [-1,  5,  6,  7, -1, -1, 14, 15, 16, -1,
                   -1, 23, 24, 25, -1, -1, -1, -1, -1, -1],
                  [-1, -1, -1, -1, -1,  9, 10, 11, 12, -1,
                   18, 19, 20, 21, -1, -1, -1, -1, -1, -1]]))


@with_setup(setup_grid)
def test_node_at_link_tail():
    assert_array_equal(
        rmg.node_at_link_tail,
        np.array([0, 1, 2, 3,
                  0, 1, 2, 3, 4,
                  5, 6, 7, 8,
                  5, 6, 7, 8, 9,
                  10, 11, 12, 13,
                  10, 11, 12, 13, 14,
                  15, 16, 17, 18]))


@with_setup(setup_grid)
def test_node_at_link_head():
    assert_array_equal(
        rmg.node_at_link_head,
        np.array([ 1,  2,  3,  4,
                   5,  6,  7,  8,  9,
                   6,  7,  8,  9,
                  10, 11, 12, 13, 14,
                  11, 12, 13, 14,
                  15, 16, 17, 18, 19,
                  16, 17, 18, 19]))


@with_setup(setup_grid)
def test_link_num_inlink():
    assert_array_equal(rmg._node_numinlink,
                       np.array([0, 1, 1, 1, 1,
                                 1, 2, 2, 2, 2,
                                 1, 2, 2, 2, 2,
                                 1, 2, 2, 2, 2]))


@with_setup(setup_grid)
def test_link_num_outlink():
    assert_array_equal(rmg._node_numoutlink, np.array([2, 2, 2, 2, 1,
                                                      2, 2, 2, 2, 1,
                                                      2, 2, 2, 2, 1,
                                                      1, 1, 1, 1, 0]))


@with_setup(setup_grid)
def test__node_inlink_matrix():
    assert_array_equal(rmg._node_inlink_matrix,
                       np.array([[-1, -1, -1, -1, -1,
                                   4,  5,  6,  7,  8,
                                  13, 14, 15, 16, 17,
                                  22, 23, 24, 25, 26],
                                 [-1,  0,  1,  2,  3,
                                  -1,  9, 10, 11, 12,
                                  -1, 18, 19, 20, 21,
                                  -1, 27, 28, 29, 30]]))


@with_setup(setup_grid)
def test__node_outlink_matrix():
    assert_array_equal(rmg._node_outlink_matrix,
                       np.array([[ 4,  5,  6,  7,  8,
                                  13, 14, 15, 16, 17,
                                  22, 23, 24, 25, 26,
                                  -1, -1, -1, -1, -1],
                                 [ 0,  1,  2,  3, -1,
                                   9, 10, 11, 12, -1,
                                  18, 19, 20, 21, -1,
                                  27, 28, 29, 30, -1]]))


@with_setup(setup_grid)
def test_links_at_node_with_scalar_interior():
    assert_array_equal(rmg.links_at_node[6],
                       np.array([10, 14, 9, 5]))


@with_setup(setup_grid)
def test_links_at_node_with_scalar_boundary():
    assert_array_equal(rmg.links_at_node[1], np.array([1,  5,  0, -1]))


@with_setup(setup_grid)
def test_links_at_node_with_array_arg():
    assert_array_equal(rmg.links_at_node[6:8],
                       np.array([[10, 14,  9,  5], [11, 15, 10,  6]]))


@with_setup(setup_grid)
def test_links_at_node_with_no_args():
    assert_array_equal(
        rmg.links_at_node,
        np.array([[ 0,  4, -1, -1],
               [ 1,  5,  0, -1],
               [ 2,  6,  1, -1],
               [ 3,  7,  2, -1],
               [-1,  8,  3, -1],
               [ 9, 13, -1,  4],
               [10, 14,  9,  5],
               [11, 15, 10,  6],
               [12, 16, 11,  7],
               [-1, 17, 12,  8],
               [18, 22, -1, 13],
               [19, 23, 18, 14],
               [20, 24, 19, 15],
               [21, 25, 20, 16],
               [-1, 26, 21, 17],
               [27, -1, -1, 22],
               [28, -1, 27, 23],
               [29, -1, 28, 24],
               [30, -1, 29, 25],
               [-1, -1, 30, 26]]))

@with_setup(setup_grid)
def test_face_at_link():
    assert_array_equal(rmg.face_at_link,
                       np.array([X, X, X, X,
                                 X, 0, 1, 2, X,
                                 3, 4, 5, 6,
                                 X, 7, 8, 9, X,
                                 10, 11, 12, 13,
                                 X, 14, 15, 16, X,
                                 X, X, X, X]))


@with_setup(setup_grid)
def test_grid_coords_to_node_id_with_scalar():
    assert_equal(rmg.grid_coords_to_node_id(3, 4), 19)


@with_setup(setup_grid)
def test_grid_coords_to_node_id_with_array():
    assert_array_equal(rmg.grid_coords_to_node_id((3, 2), (4, 1)),
                       np.array([19, 11]))


@with_setup(setup_grid)
def test_grid_coords_to_node_id_outside_of_grid():
    assert_raises(ValueError, rmg.grid_coords_to_node_id, 5, 0)


@with_setup(setup_grid)
def test_create_diagonal_list():
    rmg._create_diagonal_list()

    assert_array_equal(
        rmg._get_diagonal_list(),
        np.array([[6, X, X, X], [7, 5, X, X], [8, 6, X, X],
                  [9, 7, X, X], [X, 8, X, X],
                  [11, X, X, 1], [12, 10,  0,  2], [13, 11,  1,  3],
                  [14, 12,  2,  4], [X, 13, 3, X],
                  [16, X, X, 6], [17, 15,  5,  7], [18, 16,  6,  8],
                  [19, 17,  7,  9], [X, 18, 8, X],
                  [X, X, X, 11], [X, X, 10, 12], [X, X, 11, 13],
                  [X, X, 12, 14], [X, X, 13, X]]))



"""Generate precipitation using the Poisson pulse model.

Landlab component that generates precipitation events using the rectangular
Poisson pulse model described in Eagleson (1978, Water Resources Research).


No particular units must be used, but it was written with the storm units in
hours (hr) and depth units in millimeters (mm)

Written by Jordan Adams, 2013, updated May 2016
"""


import random
import numpy as np
from landlab import Component


class PrecipitationDistribution(object):

    """Generate precipitation events.

    This component can generate a random storm duration, interstorm
    duration, precipitation intensity or storm depth from a Poisson
    distribution when given a mean value.

    Construction::

        PrecipitationDistribution(mean_storm_duration=0.0,
                                  mean_interstorm_duration=0.0,
                                  mean_storm_depth=0.0, total_t=0.0,
                                  delta_t=0.0)

    Parameters
    ----------
    mean_storm_duration : float
        Average duration of a precipitation event.
    mean_interstorm_duration : float
        Average duration between precipitation events.
    mean_storm_depth : float
        Average depth of precipitation events.
    total_t : float, optional
        If generating a time series, the total amount of time .
    delta_t : float, optional
        If you want to break up storms into determined subsections using
        yield_storm_interstorm_duration_intensity, a delta_t is needed.

    Examples
    --------
    >>> from landlab.components.uniform_precip import PrecipitationDistribution

    To use hard-coded values for mean storm, mean interstorm, mean depth,
    model run time and delta t...  Say we use 1.5 for mean storm, 15 for mean
    interstorm, 0.5 for mean depth, 100 for model run time and 1 for delta t...

    >>> precip = PrecipitationDistribution(mean_storm_duration = 1.5,
    ...     mean_interstorm_duration = 15.0, mean_storm_depth = 0.5,
    ...     total_t = 100.0, delta_t = 1)
    """

    def __init__(self, mean_storm_duration=0.0, mean_interstorm_duration=0.0,
                 mean_storm_depth=0.0, total_t=0.0, delta_t=0.0, **kwds):
        """Create the storm generator.

        Parameters
        ----------
        mean_storm_duration : float
            Average duration of a precipitation event.
        mean_interstorm_duration : float
            Average duration between precipitation events.
        mean_storm_depth : float
            Average depth of precipitation events.
        total_t : float, optional
            If generating a time series, the total amount of time .
        delta_t : float, optional
            If you want to break up storms into determined subsections using
            yield_storm_interstorm_duration_intensity, a delta_t is needed.
        """

        self.mean_storm_duration = mean_storm_duration

        self.mean_interstorm_duration = mean_interstorm_duration

        self.mean_storm_depth = mean_storm_depth

        self.run_time = total_t

        self.delta_t = delta_t

        # Mean_intensity is not set by the MPD, but can be drawn from
        # the mean storm depth and mean storm duration.
        self.mean_intensity = self.mean_storm_depth / self.mean_storm_duration

        # If a time series is created later, this blank list will be used.
        self.storm_time_series = []

        # Given the mean values assigned above using either the model
        # parameter dictionary or the init function, we can call the
        # different methods to assign values from the Poisson distribution.

        self.storm_duration = self.get_precipitation_event_duration()
        self.interstorm_duration = self.get_interstorm_event_duration()
        self.storm_depth = self.get_storm_depth()
        self.intensity = self.get_storm_intensity()
        self._elapsed_time = 0.

    def update(self):
        """Update the storm values.

        If new values for storm duration, interstorm duration, storm depth
        and intensity are needed, this method can be used to update those
        values one time.

        Examples
        --------
        >>> from landlab.components import PrecipitationDistribution
        >>> precip = PrecipitationDistribution(mean_storm_duration=1.5,
        ...     mean_interstorm_duration=15.0, mean_storm_depth=0.5,
        ...     total_t=100.0, delta_t=1)

        Additionally, if we wanted to update several times, a loop could be
        utilized to accomplish this. Say we want 5 storm_durations; this
        pseudo-code represents a way to accomplish this...

        >>> storm_duration_list = []
        >>> i = 0
        >>> while i < 4:
        ...     storm_duration_list.append(precip.storm_duration)
        ...     precip.update()
        ...     i += 1
        """
        self.storm_duration = self.get_precipitation_event_duration()
        self.interstorm_duration = self.get_interstorm_event_duration()
        self.storm_depth = self.get_storm_depth()
        self.intensity = self.get_storm_intensity()

    def get_precipitation_event_duration(self):
        """This method is the storm generator.

        This method has one argument: the mean_storm_duration parameter.
        (In Eagleson (1978), this parameter was called Tr.)

        It finds a random storm_duration value
        based on the poisson distribution about the mean.
        This is accomplished using the expovariate function
        from the "random" standard library.
        Additionally, it is rounded to contain 4 significant figures,
        for neatness.

        The if-else statement is very important here. Values of 0
        can exist in the Poission distribution, but it does not make
        sense to have 0 duration storms, so to avoid that,
        we return a storm duration IF it is greater than 0,
        otherwise, recursion is employed to re-call the storm
        generator function and get a new value.

        Returns
        -------
        float
            The storm duration.
        """
        storm = round(random.expovariate(1 / self.mean_storm_duration), 2)
        while storm == 0:
            storm = round(random.expovariate(1 / self.mean_storm_duration), 2)
        self.storm_duration = storm
        return self.storm_duration

    def get_interstorm_event_duration(self):
        """Generate interstorm events.

        This method takes one argument, the mean_interstorm_duration parameter.
        (In Eagleson (1978), this parameter was called Tb.)

        This method is modeled identically to
        get_precipitation_event_duration()

        This method finds a random value for interstorm_duration
        based on the poisson distribution about the mean.
        This is accomplished using the expovariate function
        from the "random" standard library.
        Additionally, it is rounded to contain 4 significant figures, for
        neatness.

        The if-else statement is very important here. Values of 0
        can exist in the Poission distribution, but it does not make
        sense to have 0 hour interstorm durations.
        To avoid 0 hour interstorm durations, we return a
        interstorm duration IF it is greater than 0,
        otherwise, recursion is employed to re-call the interstorm
        duration generator function and get a new value.

        Returns
        -------
        float
            The interstorm duration.
        """
        interstorm = (round(random.expovariate(
                                        1 / self.mean_interstorm_duration), 2))
        while interstorm == 0:
            interstorm = (round(random.expovariate(
                                        1 / self.mean_interstorm_duration), 2))
        self.interstorm_duration = interstorm
        return self.interstorm_duration

    def get_storm_depth(self):
        """Generate storm depth.

        Storm depth is used to generate a realistic
        intensity for different storm events.

        (In Eagleson (1978) this parameter was called "h")

        This method requires storm_duration, mean_storm_duration
        and the mean_storm_depth. Storm_duration is generated through
        the initialize() or update() method. mean_storm_duration and
        mean_storm_depth
        are read in using the ModelParameterDictionary.

        Numpy has a random number generator to get values
        from a given Gamma distribution. It takes two arguments,
        alpha (or the shape parameter), which is the generated over the mean
        event and beta (or the scale parameter), which is the mean value
        These are all arguments in the function, which returns storm depth.

        Returns
        -------
        float
            The storm depth.
        """

        shape_parameter = (self.storm_duration / self.mean_storm_duration)
        scale_parameter = (self.mean_storm_depth)
        self.storm_depth = np.random.gamma(shape_parameter, scale_parameter)
        return self.storm_depth

    def get_storm_intensity(self):
        """Get the storm intensity.

        This method draws storm intensity out of the storm depth generated by
        get_storm_depth.

        This method requires the storm_depth and storm_duration
        and is the same as the parameter ("i") in Eagleson (1978), but instead
        of being drawn from Poission, this is drawn from the Gamma distribution
        of (*h*), as :math:`h = i * Tr`.

        Returns
        -------
        float
            The storm intensity.
        """
        self.intensity = self.storm_depth / self.storm_duration
        return self.intensity

    def get_storm_time_series(self):
        """Get a time series of storms.

        This method creates a time series of storms based on storm_duration,
        and interstorm_duration. From these values it will calculate a complete
        time series.

        The storm_time_series returned by this method is made up of sublists,
        each comprising of three sub-parts (e.g. [[x,y,z], [a,b,c]]) where x
        and a are the beginning times of a precipitation event, y and b are the
        ending times of the precipitation event and z and c represent the
        average intensity (mm/hr) of the storm lasting from x to y and a to be,
        respectively.

        Retruns
        -------
        array
            containing several sub-arrays of events [start, finish, intensity]
        """

        storm = self.get_precipitation_event_duration()
        self.get_storm_depth()
        intensity = self.get_storm_intensity()
        self.storm_time_series.append([0, storm, intensity])

        storm_helper = storm
        storm_iterator = storm
        while storm_iterator <= self.run_time:
            next_storm_start = storm_helper + (round(
                                    self.get_interstorm_event_duration(), 2))
            next_storm_end = next_storm_start + (round(
                                self.get_precipitation_event_duration(), 2))
            intensity = round(self.get_storm_intensity(), 2)
            self.get_storm_depth()
            self.storm_time_series.append(
                            [next_storm_start, next_storm_end, intensity])
            storm_iterator = storm_helper
            storm_helper = next_storm_end
            storm_iterator = storm_helper
        return self.storm_time_series

    def yield_storm_interstorm_duration_intensity(self,
                                                  subdivide_interstorms=False):
        """Iterator for a time series of storms.

        This method is intended to be equivalent to get_storm_time_series,
        but instead offers a generator functionality. This will be useful in
        cases where the whole sequence of storms and interstorms doesn't need
        to be stored, where we can save memory this way.

        The method keeps track of the delta_t such that if a storm needs to be
        generated longer than this supplied model timestep, the generator will
        return the storm in "chunks", until there is no more storm duration.
        e.g.,
        storm of intensity 1. is 4.5 long, the delta_t is 2., the generator
        yields (2.,1.) -> (2.,1.) -> (0.5,1.) -> ...

        If delta_t is None or not supplied, no subdivision occurs.

        Once a storm has been generated, this method will follow it with the
        next interstorm, yielded as (interstorm_duration, 0.). Note that the
        interstorm will NOT be subdivided according to delta_t unless you set
        the flag *subdivide_interstorms* to True.

        The method will keep yielding until it reaches the RUN_TIME, where it
        will terminate.

        Yields
        ------
        tuple of float
            (interval_duration, rainfall_rate_in_interval)

        Notes
        -----
        One recommended procedure is to instantiate the generator, then call
        instance.next() repeatedly to get the sequence.
        """
        # Added DEJH, Dec 2014
        delta_t = self.delta_t
        if delta_t is None:
            assert subdivide_interstorms is False, (
                'You specified you wanted storm subdivision, but did not ' +
                'provide a delta_t to allow this!')
        self._elapsed_time = 0.
        while self._elapsed_time < self.run_time:
            storm_duration = self.get_precipitation_event_duration()
            step_time = 0.
            self.get_storm_depth()
            intensity = self.get_storm_intensity()  # this is a rainfall rate
            if self._elapsed_time + storm_duration > self.run_time:
                storm_duration = self.run_time - self._elapsed_time
            while delta_t is not None and storm_duration - step_time > delta_t:
                yield (delta_t, intensity)
                step_time += delta_t
            yield (storm_duration - step_time, intensity)
            self._elapsed_time += storm_duration

            interstorm_duration = self.get_interstorm_event_duration()
            if self._elapsed_time + interstorm_duration > self.run_time:
                interstorm_duration = self.run_time - self._elapsed_time
            if subdivide_interstorms:
                step_time = 0.
                while interstorm_duration-step_time > delta_t:
                    yield (delta_t, 0.)
                    step_time += delta_t
                yield (interstorm_duration - step_time, 0.)
            else:
                yield (interstorm_duration, 0.)
            self._elapsed_time += interstorm_duration

    @property
    def elapsed_time(self):
        """Get the elapsed time recorded by the module.

        This will be particularly useful in the midst of a yield loop.
        """
        return self._elapsed_time



from .chi_index import ChiFinder
from .diffusion import LinearDiffuser
from .fire_generator import FireGenerator
from .detachment_ltd_erosion import DetachmentLtdErosion
from .flexure import Flexure
from .flow_accum import AccumFlow
from .flow_routing import FlowRouter, DepressionFinderAndRouter
from .nonlinear_diffusion import PerronNLDiffuse
from .overland_flow import OverlandFlowBates, OverlandFlow
from .potentiality_flowrouting import PotentialityFlowRouter
from .sink_fill import SinkFiller
from .steepness_index import SteepnessFinder
from .stream_power import StreamPowerEroder, FastscapeEroder, SedDepEroder
from .uniform_precip import PrecipitationDistribution
from .gflex import gFlex
# from .vegetation_ca import VegCA
# from .soil_moisture import SoilMoisture
from .radiation import Radiation
# from .single_vegetation import Vegetation
from .pet import PotentialEvapotranspiration


COMPONENTS = [ChiFinder, LinearDiffuser,
              Flexure, AccumFlow, FlowRouter, DepressionFinderAndRouter,
              PerronNLDiffuse, OverlandFlowBates, OverlandFlow,
              PotentialEvapotranspiration,PotentialityFlowRouter,
              Radiation, SinkFiller, StreamPowerEroder,
              FastscapeEroder, SedDepEroder,
              SteepnessFinder, DetachmentLtdErosion, gFlex]


__all__ = [cls.__name__ for cls in COMPONENTS]



#! /usr/env/python
"""

Component that models 2D diffusion using an explicit finite-volume method.

Created July 2013 GT
Last updated March 2016 DEJH with LL v1.0 component style
"""

from __future__ import print_function

import numpy as np
from six.moves import range

from landlab import ModelParameterDictionary, Component, FieldError, \
    create_and_initialize_grid, FIXED_GRADIENT_BOUNDARY, FIXED_LINK
from landlab.core.model_parameter_dictionary import MissingKeyError
from landlab.utils.decorators import use_file_name_or_kwds

_ALPHA = 0.15   # time-step stability factor
# ^0.25 not restrictive enough at meter scales w S~1 (possible cases)


class LinearDiffuser(Component):
    """
    This component implements linear diffusion of a Landlab field.

    Component assumes grid does not deform. If the boundary conditions on the
    grid change after component instantiation, be sure to also call
    :func:`updated_boundary_conditions` to ensure these are reflected in the
    component (especially if fixed_links are present).

    The primary method of this class is :func:`run_one_step`.

    Construction::

        LinearDiffuser(grid, linear_diffusivity=None)

    Parameters
    ----------
    grid : ModelGrid
        A grid.
    linear_diffusivity : float, array, or field name (m**2/time)
        The diffusivity.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> import numpy as np
    >>> mg = RasterModelGrid((9, 9), 1.)
    >>> z = mg.add_zeros('node', 'topographic__elevation')
    >>> z.reshape((9, 9))[4, 4] = 1.
    >>> mg.set_closed_boundaries_at_grid_edges(True, True, True, True)
    >>> ld = LinearDiffuser(mg, linear_diffusivity=1.)
    >>> for i in range(1):
    ...     ld.run_one_step(1.)
    >>> np.isclose(z[mg.core_nodes].sum(), 1.)
    True
    >>> mg2 = RasterModelGrid((5, 30), 1.)
    >>> z2 = mg2.add_zeros('node', 'topographic__elevation')
    >>> z2.reshape((5, 30))[2, 8] = 1.
    >>> z2.reshape((5, 30))[2, 22] = 1.
    >>> mg2.set_closed_boundaries_at_grid_edges(True, True, True, True)
    >>> kd = mg2.node_x/mg2.node_x.mean()
    >>> ld2 = LinearDiffuser(mg2, linear_diffusivity=kd)
    >>> for i in range(10):
    ...     ld2.run_one_step(0.1)
    >>> z2[mg2.core_nodes].sum() == 2.
    True
    >>> z2.reshape((5, 30))[2, 8] > z2.reshape((5, 30))[2, 22]
    True
    """

    _name = 'LinearDiffuser'

    _input_var_names = ('topographic__elevation',)

    _output_var_names = ('topographic__elevation',
                         'topographic__gradient',
                         'unit_flux',
                         )

    _var_units = {'topographic__elevation': 'm',
                  'topographic__gradient': '-',
                  'unit_flux': 'm**3/s',
                  }

    _var_mapping = {'topographic__elevation': 'node',
                    'topographic__gradient': 'link',
                    'unit_flux': 'link',
                    }

    _var_doc = {
        'topographic__elevation': ('Land surface topographic elevation; can ' +
                                   'be overwritten in initialization'),
        'topographic__gradient': 'Gradient of surface, on links',
        'unit_flux': 'Volume flux per unit width along links'
    }

    @use_file_name_or_kwds
    def __init__(self, grid, linear_diffusivity=None, **kwds):
        self._grid = grid
        self.current_time = 0.
        if linear_diffusivity is not None:
            if type(linear_diffusivity) is not str:
                self.kd = linear_diffusivity
            else:
                self.kd = self.grid.at_node[linear_diffusivity]
        else:
            raise KeyError("linear_diffusivity must be provided to the " +
                           "LinearDiffuser component")

        # for component back compatibility (undocumented):
        # note component can NO LONGER do internal uplift, at all.
        # ###
        self.timestep_in = kwds.pop('dt', None)
        if 'values_to_diffuse' in kwds.keys():
            self.values_to_diffuse = kwds.pop('values_to_diffuse')
            for mytups in (self._input_var_names, self._output_var_names):
                myset = set(mytups)
                myset.remove('topographic__elevation')
                myset.add(self.values_to_diffuse)
                mytups = tuple(myset)
            for mydicts in (self._var_units, self._var_mapping, self._var_doc):
                mydicts[self.values_to_diffuse] = mydicts.pop(
                    'topographic__elevation')
        else:
            self.values_to_diffuse = 'topographic__elevation'
        # Raise an error if somehow someone is using this weird functionality
        if self._grid is None:
            raise ValueError('You must now provide an existing grid!')
        # ###

        # Set internal time step
        # ..todo:
        #   implement mechanism to compute time-steps dynamically if grid is
        #   adaptive/changing
        # as of modern componentization (Spring '16), this can take arrays
        # and irregular grids
        if type(self.kd) not in (int, float):
            assert len(self.kd) == self.grid.number_of_nodes, self.kd
            kd_links = self.grid.map_max_of_link_nodes_to_link(self.kd)
        else:
            kd_links = float(self.kd)
        # assert CFL condition:
        CFL_prefactor = _ALPHA * self.grid.length_of_link ** 2.
        self._CFL_actives_prefactor = CFL_prefactor[self.grid.active_links]
        # ^note we can do this as topology shouldn't be changing
        dt_links = CFL_prefactor / kd_links
        self.dt = np.amin(dt_links[self.grid.active_links])

        # Get a list of interior cells
        self.interior_cells = self.grid.node_at_core_cell

        self.z = self.grid.at_node[self.values_to_diffuse]
        g = self.grid.zeros(at='link')
        qs = self.grid.zeros(at='link')
        try:
            self.g = self.grid.add_field('link', 'topographic__gradient', g,
                                         noclobber=True)
            # ^note this will object if this exists already
        except FieldError:
            self.g = self.grid.at_link['topographic__gradient']  # keep a ref
        try:
            self.qs = self.grid.add_field('link', 'unit_flux', qs,
                                          noclobber=True)
        except FieldError:
            self.qs = self.grid.at_link['unit_flux']
        # note all these terms are deliberately loose, as we won't always be
        # dealing with topo

        # do some pre-work to make fixed grad BC updating faster in the loop:
        self.updated_boundary_conditions()

    def updated_boundary_conditions(self):
        """Call if grid BCs are updated after component instantiation.

        Sets `fixed_grad_nodes`, `fixed_grad_anchors`, & `fixed_grad_offsets`,
        such that::

            value[fixed_grad_nodes] = value[fixed_grad_anchors] + offset

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> import numpy as np
        >>> mg = RasterModelGrid((4, 5), 1.)
        >>> z = mg.add_zeros('node', 'topographic__elevation')
        >>> z[mg.core_nodes] = 1.
        >>> ld = LinearDiffuser(mg, linear_diffusivity=1.)
        >>> ld.fixed_grad_nodes.size == 0
        True
        >>> ld.fixed_grad_anchors.size == 0
        True
        >>> ld.fixed_grad_offsets.size == 0
        True
        >>> mg.at_link['topographic__slope'] = mg.calc_grad_at_link(
        ...     'topographic__elevation')
        >>> mg.set_fixed_link_boundaries_at_grid_edges(True, True, True, True)
        >>> ld.updated_boundary_conditions()
        >>> ld.fixed_grad_nodes
        array([ 1,  2,  3,  5,  9, 10, 14, 16, 17, 18])
        >>> ld.fixed_grad_anchors
        array([ 6,  7,  8,  6,  8, 11, 13, 11, 12, 13])
        >>> ld.fixed_grad_offsets
        array([-1., -1., -1., -1., -1., -1., -1., -1., -1., -1.])
        >>> np.allclose(z[ld.fixed_grad_nodes],
        ...             z[ld.fixed_grad_anchors] + ld.fixed_grad_offsets)
        True
        """
        fixed_grad_nodes = np.where(self.grid.status_at_node ==
                                    FIXED_GRADIENT_BOUNDARY)[0]
        heads = self.grid.node_at_link_head[self.grid.fixed_links]
        tails = self.grid.node_at_link_tail[self.grid.fixed_links]
        head_is_fixed = np.in1d(heads, fixed_grad_nodes)
        self.fixed_grad_nodes = np.where(head_is_fixed, heads, tails)
        self.fixed_grad_anchors = np.where(head_is_fixed, tails, heads)
        vals = self.grid.at_node[self.values_to_diffuse]
        self.fixed_grad_offsets = (vals[self.fixed_grad_nodes] -
                                   vals[self.fixed_grad_anchors])

    def diffuse(self, dt, **kwds):
        """
        See :func:`run_one_step`.
        """
        if 'internal_uplift' in kwds.keys():
            raise KeyError('LinearDiffuser can no longer work with internal ' +
                           'uplift')
        # Take the smaller of delt or built-in time-step size self.dt
        self.tstep_ratio = dt/self.dt
        repeats = int(self.tstep_ratio//1.)
        extra_time = self.tstep_ratio - repeats
        z = self.grid.at_node[self.values_to_diffuse]

        core_nodes = self.grid.node_at_core_cell
        # do mapping of array kd here, in case it points at an updating field:
        if type(self.kd) is np.ndarray:
            kd_activelinks = self.grid.map_max_of_link_nodes_to_link(
                self.kd)[self.grid.active_links]
            # re-derive CFL condition, as could change dynamically:
            dt_links = self._CFL_actives_prefactor / kd_activelinks
            self.dt = dt_links.min()
        else:
            kd_activelinks = self.kd

        for i in range(repeats+1):
            # Calculate the gradients and sediment fluxes
            self.g[self.grid.active_links] = \
                    self.grid.calc_grad_at_link(z)[self.grid.active_links]
            # if diffusivity is an array, self.kd is already active_links-long
            self.qs[self.grid.active_links] = (-kd_activelinks *
                                               self.g[self.grid.active_links])

            # Calculate the net deposition/erosion rate at each node
            self.dqsds = self.grid.calc_flux_div_at_node(self.qs)
            # Calculate the total rate of elevation change
            dzdt = - self.dqsds

            # Update the elevations
            timestep = self.dt
            if i == (repeats):
                timestep *= extra_time
            else:
                pass
            self.grid.at_node[self.values_to_diffuse][core_nodes] += dzdt[
                core_nodes] * timestep

            # check the BCs, update if fixed gradient
            vals = self.grid.at_node[self.values_to_diffuse]
            vals[self.fixed_grad_nodes] = (vals[self.fixed_grad_anchors] +
                                           self.fixed_grad_offsets)

        return self.grid

    def run_one_step(self, dt, **kwds):
        """Run the diffuser for one timestep, dt.

        If the imposed timestep dt is longer than the Courant-Friedrichs-Lewy
        condition for the diffusion, this timestep will be internally divided
        as the component runs, as needed.

        Parameters
        ----------
        dt : float (time)
            The imposed timestep.
        """
        self.diffuse(dt, **kwds)

    @property
    def time_step(self):
        """Returns internal time-step size (as a property).
        """
        return self.dt



#! /usr/env/python

"""
This module attempts to "component-ify" GT's Fastscape stream power erosion.
Created DEJH, March 2014.
"""
from __future__ import print_function

import numpy
import warnings
from landlab import ModelParameterDictionary, Component
from landlab.core.model_parameter_dictionary import MissingKeyError, \
    ParameterValueError
from landlab.utils.decorators import use_file_name_or_kwds
from landlab.field.scalar_data_fields import FieldError
from scipy.optimize import newton, fsolve

UNDEFINED_INDEX = -1


class FastscapeEroder(Component):
    '''
    This class uses the Braun-Willett Fastscape approach to calculate the
    amount of erosion at each node in a grid, following a stream power
    framework. This should allow it to be stable against larger timesteps
    than an explicit stream power scheme.

    Stream power erosion is implemented as::

        E = K * (rainfall_intensity*A)**m * S**n - threshold_sp,

    if K * A**m * S**n > threshold_sp, and::

        E = 0,

    if K * A**m * S**n <= threshold_sp.

    This module assumes you have already run
    :func:`landlab.components.flow_routing.route_flow_dn.FlowRouter.route_flow`
    in the same timestep. It looks for 'flow__upstream_node_order',
    'flow__link_to_receiver_node', 'drainage_area', 'flow__receiver_node', and
    'topographic__elevation' at the nodes in the grid. 'drainage_area' should
    be in area upstream, not volume (i.e., set runoff_rate=1.0 when calling
    FlowRouter.route_flow).

    The primary method of this class is :func:`run_one_step`.

    Construction::

        FastscapeEroder(grid, K_sp=None, m_sp=0.5, n_sp=1., threshold_sp=0.,
                        rainfall_intensity=1.)

    Parameters
    ----------
    grid : ModelGrid
        A grid.
    K_sp : float, array, or field name
        K in the stream power equation (units vary with other parameters).
    m_sp : float, optional
        m in the stream power equation (power on drainage area).
    n_sp : float, optional, ~ 0.5<n_sp<4.
        n in the stream power equation (power on slope).
        Performance will be VERY degraded if n < 1.
    threshold_sp : float, array, or field name
        The threshold stream power.
    rainfall intensity : float, array, or field name; optional
        Modifying factor on drainage area to convert it to a true water
        volume flux in (m/time). i.e., E = K * (r_i*A)**m * S**n

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> from landlab import CLOSED_BOUNDARY, FIXED_VALUE_BOUNDARY
    >>> from landlab.components import FlowRouter
    >>> mg = RasterModelGrid((5, 5), 10.)
    >>> z = np.array([7.,  7.,  7.,  7.,  7.,
    ...               7.,  5., 3.2,  6.,  7.,
    ...               7.,  2.,  3.,  5.,  7.,
    ...               7.,  1., 1.9,  4.,  7.,
    ...               7.,  0.,  7.,  7.,  7.])
    >>> z = mg.add_field('node', 'topographic__elevation', z)
    >>> fr = FlowRouter(mg)
    >>> sp = FastscapeEroder(mg, K_sp=1.)
    >>> fr.run_one_step()
    >>> sp.run_one_step(dt=1.)
    >>> z  # doctest: +NORMALIZE_WHITESPACE
    array([ 7.        ,  7.        ,  7.        ,  7.        ,  7.        ,
            7.        ,  2.92996598,  2.02996598,  4.01498299,  7.        ,
            7.        ,  0.85993197,  1.87743897,  3.28268321,  7.        ,
            7.        ,  0.28989795,  0.85403051,  2.42701526,  7.        ,
            7.        ,  0.        ,  7.        ,  7.        ,  7.        ])

    >>> mg2 = RasterModelGrid((3, 7), 1.)
    >>> z = np.array(mg2.node_x**2.)
    >>> z = mg2.add_field('node', 'topographic__elevation', z)
    >>> mg2.status_at_node[mg2.nodes_at_left_edge] = FIXED_VALUE_BOUNDARY
    >>> mg2.status_at_node[mg2.nodes_at_top_edge] = CLOSED_BOUNDARY
    >>> mg2.status_at_node[mg2.nodes_at_bottom_edge] = CLOSED_BOUNDARY
    >>> mg2.status_at_node[mg2.nodes_at_right_edge] = CLOSED_BOUNDARY
    >>> fr2 = FlowRouter(mg2)
    >>> sp2 = FastscapeEroder(mg2, K_sp=0.1, m_sp=0., n_sp=2.,
    ...                       threshold_sp=2.)
    >>> fr2.run_one_step()
    >>> sp2.run_one_step(dt=10.)
    >>> z.reshape((3, 7))[1, :]  # doctest: +NORMALIZE_WHITESPACE
    array([  0.        ,   1.        ,   4.        ,   8.52493781,
            13.29039716,  18.44367965,  36.        ])

    >>> mg3 = RasterModelGrid((3, 7), 1.)
    >>> z = np.array(mg3.node_x**2.)
    >>> z = mg3.add_field('node', 'topographic__elevation', z)
    >>> mg3.status_at_node[mg3.nodes_at_left_edge] = FIXED_VALUE_BOUNDARY
    >>> mg3.status_at_node[mg3.nodes_at_top_edge] = CLOSED_BOUNDARY
    >>> mg3.status_at_node[mg3.nodes_at_bottom_edge] = CLOSED_BOUNDARY
    >>> mg3.status_at_node[mg3.nodes_at_right_edge] = CLOSED_BOUNDARY
    >>> fr3 = FlowRouter(mg3)
    >>> K_field = mg3.ones('node')  # K can be a field
    >>> sp3 = FastscapeEroder(mg3, K_sp=K_field, m_sp=1., n_sp=0.6,
    ...                       threshold_sp=mg3.node_x,
    ...                       rainfall_intensity=z)
    >>> fr3.run_one_step()
    >>> sp3.run_one_step(1.)
    >>> z.reshape((3, 7))[1, :]  # doctest: +NORMALIZE_WHITESPACE
    array([  0.        ,   0.        ,   0.18508544,   0.42869679,
             0.85354205,   2.05732094,  36.        ])
    '''

    _name = 'FastscapeEroder'

    _input_var_names = (
        'topographic__elevation',
        'drainage_area',
        'flow__link_to_receiver_node',
        'flow__upstream_node_order',
        'flow__receiver_node',
    )

    _output_var_names = (
        'topographic__elevation',
    )

    _var_units = {
        'topographic__elevation': 'm',
        'drainage_area': 'm**2',
        'flow__link_to_receiver_node': '-',
        'flow__upstream_node_order': '-',
        'flow__receiver_node': '-',
    }

    _var_mapping = {
        'topographic__elevation': 'node',
        'drainage_area': 'node',
        'flow__link_to_receiver_node': 'node',
        'flow__upstream_node_order': 'node',
        'flow__receiver_node': 'node',
    }

    _var_doc = {
        'topographic__elevation': 'Land surface topographic elevation',
        'drainage_area':
            "Upstream accumulated surface area contributing to the node's "
            "discharge",
        'flow__link_to_receiver_node':
            'ID of link downstream of each node, which carries the discharge',
        'flow__upstream_node_order':
            'Node array containing downstream-to-upstream ordered list of '
            'node IDs',
        'flow__receiver_node':
            'Node array of receivers (node that receives flow from current '
            'node)',
    }

    @use_file_name_or_kwds
    def __init__(self, grid, K_sp=None, m_sp=0.5, n_sp=1., threshold_sp=0.,
                 rainfall_intensity=1., **kwds):
        """
        Initialize the Fastscape stream power component. Note: a timestep,
        dt, can no longer be supplied to this component through the input file.
        It must instead be passed directly to the run method.

        Parameters
        ----------
        grid : ModelGrid
            A grid.
        K_sp : float, array, or field name
            K in the stream power equation (units vary with other parameters).
        m_sp : float, optional
            m in the stream power equation (power on drainage area).
        n_sp : float, optional
            n in the stream power equation (power on slope).
        rainfall intensity : float, array, or field name; optional
            Modifying factor on drainage area to convert it to a true water
            volume flux in (m/time). i.e., E = K * (r_i*A)**m * S**n
        """
        self._grid = grid

        self.K = K_sp  # overwritten below in special cases
        self.m = float(m_sp)
        self.n = float(n_sp)
        if type(threshold_sp) in (float, int):
            self.thresholds = float(threshold_sp)
        else:
            if type(threshold_sp) is str:
                self.thresholds = self.grid.at_node[threshold_sp]
            else:
                self.thresholds = threshold_sp
            assert self.thresholds.size == self.grid.number_of_nodes

        # make storage variables
        self.A_to_the_m = grid.zeros(at='node')
        self.alpha = grid.empty(at='node')
        self.alpha_by_flow_link_lengthtothenless1 = numpy.empty_like(
                                                        self.alpha)

        self.grid._diagonal_links_at_node  # calc number of diagonal links

        if self.K is None:
            raise ValueError('K_sp must be set as a float, node array, or ' +
                             'field name. It was None.')

        # now handle the inputs that could be float, array or field name:
        # some support here for old-style inputs
        if type(K_sp) is str:
            if K_sp == 'array':
                self.K = None
            else:
                self.K = self._grid.at_node[K_sp]
        elif type(K_sp) in (float, int):  # a float
            self.K = float(K_sp)
        elif len(K_sp) == self.grid.number_of_nodes:
            self.K = numpy.array(K_sp)
        else:
            raise TypeError('Supplied type of K_sp ' +
                            'was not recognised, or array was ' +
                            'not nnodes long!')

        if type(rainfall_intensity) is str:
            if rainfall_intensity == 'array':
                self.r_i = None
            else:
                self.r_i = self._grid.at_node[rainfall_intensity]
        elif type(rainfall_intensity) in (float, int):  # a float
            self.r_i = float(rainfall_intensity)
        elif len(rainfall_intensity) == self.grid.number_of_nodes:
            self.r_i = numpy.array(rainfall_intensity)
        else:
            raise TypeError('Supplied type of rainfall_' +
                            'intensity was not recognised, or array was ' +
                            'not nnodes long!')

        # We now forbid changing of the field name
        if 'value_field' in kwds.keys():
            raise ValueError('This component can no longer support variable' +
                             'field names. Use "topographic__elevation".')

    def erode(self, grid_in, dt=None, K_if_used=None, flooded_nodes=None):
        """
        This method implements the stream power erosion, following the Braun-
        Willett (2013) implicit Fastscape algorithm. This should allow it to
        be stable against larger timesteps than an explicit stream power
        scheme.

        This driving method for this component is now superceded by the new,
        standardized wrapper :func:`run_one_step`, but is retained for
        back compatibility.

        Set 'K_if_used' as a field name or nnodes-long array if you set K_sp as
        'array' during initialization.

        It returns the grid, in which it will have modified the value of
        *value_field*, as specified in component initialization.

        Parameters
        ----------
        grid_in : a grid
            This is a dummy argument maintained for component back-
            compatibility. It is superceded by the copy of the grid passed
            during initialization.
        dt : float
            Time-step size. If you are calling the deprecated function
            :func:`gear_timestep`, that method will supercede any value
            supplied here.
        K_if_used : array (optional)
            Set this to an array if you set K_sp to 'array' in your input file.
        flooded_nodes : ndarray of int (optional)
            IDs of nodes that are flooded and should have no erosion. If not
            provided but flow has still been routed across depressions, erosion
            may still occur beneath the apparent water level (though will
            always still be positive).

        Returns
        -------
        grid
            A reference to the grid.
        """
        upstream_order_IDs = self._grid['node']['flow__upstream_node_order']
        z = self._grid['node']['topographic__elevation']
        defined_flow_receivers = numpy.not_equal(self._grid['node'][
            'flow__link_to_receiver_node'], UNDEFINED_INDEX)
        flow_link_lengths = self._grid._length_of_link_with_diagonals[
            self._grid['node']['flow__link_to_receiver_node'][
                defined_flow_receivers]]

        # make arrays from input the right size
        if type(self.K) is numpy.ndarray:
            K_here = self.K[defined_flow_receivers]
        else:
            K_here = self.K
        if type(self.r_i) is numpy.ndarray:
            r_i_here = self.r_i[defined_flow_receivers]
        else:
            r_i_here = self.r_i

        if dt is None:
            dt = self.dt
        assert dt is not None, ('Fastscape component could not find a dt to ' +
                                'use. Pass dt to the run_one_step() method.')

        if self.K is None:  # "old style" setting of array
            assert K_if_used is not None
            self.K = K_if_used

        numpy.power(self._grid['node']['drainage_area'], self.m,
                    out=self.A_to_the_m)
        self.alpha[defined_flow_receivers] = r_i_here**self.m * K_here * dt * \
            self.A_to_the_m[defined_flow_receivers] / flow_link_lengths

        flow_receivers = self._grid['node']['flow__receiver_node']
        n_nodes = upstream_order_IDs.size
        alpha = self.alpha

        # Handle flooded nodes, if any (no erosion there)
        if flooded_nodes is not None:
            alpha[flooded_nodes] = 0.
        else:
            reversed_flow = z < z[flow_receivers]
            # this check necessary if flow has been routed across depressions
            alpha[reversed_flow] = 0.

        self.alpha_by_flow_link_lengthtothenless1[
            defined_flow_receivers] = (alpha[defined_flow_receivers] /
                                       flow_link_lengths**(self.n - 1.))
        alpha_divided = self.alpha_by_flow_link_lengthtothenless1
        n = float(self.n)
        threshdt = self.thresholds * dt
        if type(self.thresholds) is float:
            from .cfuncs import erode_with_link_alpha_fixthresh
            erode_with_link_alpha_fixthresh(upstream_order_IDs, flow_receivers,
                                            threshdt, alpha_divided, n, z)
        else:
            from .cfuncs import erode_with_link_alpha_varthresh
            erode_with_link_alpha_varthresh(upstream_order_IDs, flow_receivers,
                                            threshdt, alpha_divided, n, z)
            # # This replicates the cython for testing:
            # for i in range(upstream_order_IDs.size):
            #     src_id = upstream_order_IDs[i]
            #     dst_id = flow_receivers[src_id]
            #     thresh = threshdt[i]
            #     if src_id != dst_id:
            #         next_z = z[src_id]
            #         prev_z = 0.
            #         while True:
            #         #for j in xrange(2):
            #             z_diff = next_z - z[dst_id]
            #             f = alpha_divided[src_id] * pow(z_diff, n - 1.)
            #             # if z_diff -> 0, pow -> nan (in reality, inf)
            #             # print (f, prev_z, next_z, z_diff, z[dst_id])
            #             next_z = next_z - ((next_z - z[src_id] + (
            #                 f*z_diff - thresh).clip(0.)) / (1. + n * f))
            #             if next_z < z[dst_id]:
            #                 next_z = z[dst_id] + 1.e-15
            #                 # ^maintain connectivity
            #             if next_z != 0.:
            #                 if (numpy.fabs((next_z - prev_z)/next_z) <
            #                     1.48e-08) or (n == 1.):
            #                     break
            #             else:
            #                 break
            #             prev_z = next_z
            #         if next_z < z[src_id]:
            #             z[src_id] = next_z

        return self._grid

    def run_one_step(self, dt, flooded_nodes=None, **kwds):
        """
        This method implements the stream power erosion across one time
        interval, dt, following the Braun-Willett (2013) implicit Fastscape
        algorithm.

        This follows Landlab standardized component design, and supercedes the
        old driving method :func:`erode`.

        Parameters
        ----------
        dt : float
            Time-step size
        flooded_nodes : ndarray of int (optional)
            IDs of nodes that are flooded and should have no erosion. If not
            provided but flow has still been routed across depressions, erosion
            may still occur beneath the apparent water level (though will
            always still be positive).
        """
        self.erode(grid_in=self._grid, dt=dt, flooded_nodes=flooded_nodes)



#! /usr/bin/env python
"""Some utilities for the landlab package.

Landlab utilities
+++++++++++++++++

.. autosummary::
    :toctree: generated/

    ~landlab.core.utils.radians_to_degrees
    ~landlab.core.utils.extend_array
    ~landlab.core.utils.as_id_array
    ~landlab.core.utils.make_optional_arg_into_id_array
    ~landlab.core.utils.get_functions_from_module
    ~landlab.core.utils.add_functions_to_class
    ~landlab.core.utils.add_module_functions_to_class
    ~landlab.core.utils.strip_grid_from_method_docstring
    ~landlab.core.utils.argsort_points_by_x_then_y
    ~landlab.core.utils.sort_points_by_x_then_y
    ~landlab.core.utils.anticlockwise_argsort_points
"""

import numpy as np



SIZEOF_INT = np.dtype(np.int).itemsize


def radians_to_degrees(rads):
    """Convert radians to compass-style degrees.

    Convert angles (measured counter-clockwise from the positive x-axis) in
    radians to angles in degrees measured clockwise starting from north.

    Parameters
    ----------
    rads : float or ndarray
        Angles in radians.

    Returns
    -------
    degrees : float or ndarray
        Converted angles in degrees.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.core.utils import radians_to_degrees

    >>> radians_to_degrees(0.)
    90.0
    >>> radians_to_degrees(np.pi / 2.)
    0.0
    >>> radians_to_degrees(- 3 * np.pi / 2.)
    0.0
    >>> radians_to_degrees(np.array([- np.pi, np.pi]))
    array([ 270.,  270.])
    """
    degrees = (5. * np.pi / 2. - rads) % (2. * np.pi)
    return 180. / np.pi * degrees


def extend_array(x, fill=0):
    """Extend an array by one element.

    The new array will appear as a view of the input array. However, its
    data now points to a newly-allocated buffer that's one element longer
    and contains a copy of the contents of *x*. The last element of the
    buffer is filled with *fill*. To access the extended array, use the
    *x* attribute of the returned array.

    Parameters
    ----------
    x : ndarray
        The array to extend.
    fill : number, optional
        The value to fill the extra element with.

    Returns
    -------
    ndarray
        A view of the original array with the data buffer extended by one
        element.

    Examples
    --------
    >>> from landlab.core.utils import extend_array
    >>> import numpy as np
    >>> arr = extend_array(np.arange(4).reshape((2, 2)))
    >>> arr
    array([[0, 1],
           [2, 3]])
    >>> arr.ext
    array([0, 1, 2, 3, 0])

    If the array is already extended, don't extend it more. However,
    possibly change its fill value.

    >>> rtn = extend_array(arr, fill=999)
    >>> rtn is arr
    True
    >>> rtn.ext
    array([  0,   1,   2,   3, 999])
    """
    if hasattr(x, 'ext'):
        x.ext[-1] = fill
        return x

    extended = np.empty(x.size + 1, dtype=x.dtype)
    extended[:-1] = x.flat
    extended[-1] = fill
    x.data = extended.data

    class array(np.ndarray):
        def __new__(cls, arr):
            """Instantiate the class with a view of the base array."""
            obj = np.asarray(arr).view(cls)
            obj.ext = extended
            return obj

    return array(x)


def as_id_array(array):
    """Convert an array to an array of ids.

    Parameters
    ----------
    array : ndarray
        Array of IDs.

    Returns
    -------
    ndarray
        A, possibly new, array of IDs.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.core.utils import as_id_array
    >>> x = np.arange(5)
    >>> y = as_id_array(x)
    >>> y
    array([0, 1, 2, 3, 4])

    >>> x = np.arange(5, dtype=np.int)
    >>> y = as_id_array(x)
    >>> y
    array([0, 1, 2, 3, 4])

    >>> x = np.arange(5, dtype=np.int32)
    >>> y = as_id_array(x)
    >>> y
    array([0, 1, 2, 3, 4])
    >>> y.dtype == np.int
    True

    >>> x = np.arange(5, dtype=np.int64)
    >>> y = as_id_array(x)
    >>> y
    array([0, 1, 2, 3, 4])
    >>> y.dtype == np.int
    True

    >>> x = np.arange(5, dtype=np.intp)
    >>> y = as_id_array(x)
    >>> y
    array([0, 1, 2, 3, 4])
    >>> y.dtype == np.int
    True
    """
    if array.dtype == np.int:
        return array.view(np.int)
    else:
        return array.astype(np.int)


if np.dtype(np.intp) == np.int:
    def _as_id_array(array):
        if array.dtype == np.intp or array.dtype == np.int:
            return array.view(np.int)
        else:
            return array.astype(np.int)
else:
    def _as_id_array(array):
        if array.dtype == np.int:
            return array.view(np.int)
        else:
            return array.astype(np.int)


def make_optional_arg_into_id_array(number_of_elements, *args):
    """Transform an optional argument into an array of element ids.

    Many landlab functions an optional argument of element ids that tells the
    function to operate only on the elements provided. However, if the argument
    is absent, all of the elements are to be operated on. This is a convenience
    function that converts such an arguments list into an array of elements
    ids.

    Parameters
    ----------
    number_of_elements : int
        Number of elements in the grid.
    array : array_like
        Iterable to convert to an array.

    Returns
    -------
    ndarray
        Input array converted to a numpy array, or a newly-created numpy
        array.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.core.utils import make_optional_arg_into_id_array
    >>> make_optional_arg_into_id_array(4)
    array([0, 1, 2, 3])
    >>> make_optional_arg_into_id_array(4, [0, 0, 0, 0])
    array([0, 0, 0, 0])
    >>> make_optional_arg_into_id_array(4, (1, 1, 1, 1))
    array([1, 1, 1, 1])
    >>> make_optional_arg_into_id_array(4, np.ones(4))
    array([1, 1, 1, 1])
    >>> make_optional_arg_into_id_array(4, 0)
    array([0])
    >>> make_optional_arg_into_id_array(4, np.array([[1, 2], [3, 4]]))
    array([1, 2, 3, 4])
    """
    if len(args) == 0:
        ids = np.arange(number_of_elements, dtype=np.int)
    elif len(args) == 1:
        ids = as_id_array(np.asarray(args[0])).reshape((-1, ))
    else:
        raise ValueError('Number of arguments must be 0 or 1.')

    return ids


def get_functions_from_module(mod, pattern=None):
    """Get all the function in a module.

    Parameters
    ----------
    mod : module
        An instance of a module.
    pattern : str, optional
        Only get functions whose name match a regular expression.

    Returns
    -------
    dict
        Dictionary of functions contained in the module. Keys are the
        function names, values are the functions themselves.
    """
    import inspect
    import re

    funcs = {}
    for name, func in inspect.getmembers(mod, inspect.isroutine):
        if pattern is None or re.match(pattern, name):
            funcs[name] = func
    return funcs


def add_functions_to_class(cls, funcs):
    """Add functions as methods of a class.

    Parameters
    ----------
    cls : class
        A class.
    funcs : dict
        Dictionary of function names and instances.
    """
    for name, func in funcs.items():
        setattr(cls, name, func)


def add_module_functions_to_class(cls, module, pattern=None):
    """Add functions from a module to a class as methods.

    Parameters
    ----------
    cls : class
        A class.
    module : module
        An instance of a module.
    pattern : str, optional
        Only get functions whose name match a regular expression.
    """
    import inspect
    import imp
    import os
    
    caller = inspect.stack()[1]
    path = os.path.join(os.path.dirname(caller[1]), os.path.dirname(module))

    (module, _) = os.path.splitext(os.path.basename(module))

    mod = imp.load_module(module, *imp.find_module(module, [path]))

    funcs = get_functions_from_module(mod, pattern=pattern)
    strip_grid_from_method_docstring(funcs)
    add_functions_to_class(cls, funcs)


def strip_grid_from_method_docstring(funcs):
    """Remove 'grid' from the parameters of a dict of functions' docstrings.

    Note that the docstring must be close to numpydoc standards for this to
    work.

    Parameters
    ----------
    funcs : dict
        Dictionary of functions to modify. Keys are the function names,
        values are the functions themselves.

    Examples
    --------
    >>> from landlab.grid.mappers import dummy_func_to_demonstrate_docstring_modification as dummy_func
    >>> funcs = {'dummy_func_to_demonstrate_docstring_modification':
    ...          dummy_func}
    >>> help(dummy_func)
    Help on function dummy_func_to_demonstrate_docstring_modification in module landlab.grid.mappers:
    <BLANKLINE>
    dummy_func_to_demonstrate_docstring_modification(grid, some_arg)
        A dummy function to demonstrate automated docstring changes.
    <BLANKLINE>
        Construction::
    <BLANKLINE>
            dummy_func_to_demonstrate_docstring_modification(grid, some_arg)
    <BLANKLINE>
        Parameters
        ----------
        grid : ModelGrid
            A Landlab modelgrid.
        some_arg : whatever
            A dummy argument.
    <BLANKLINE>
        Examples
        --------
        ...
    <BLANKLINE>
    >>> strip_grid_from_method_docstring(funcs)
    >>> help(dummy_func)
    Help on function dummy_func_to_demonstrate_docstring_modification in module landlab.grid.mappers:
    <BLANKLINE>
    dummy_func_to_demonstrate_docstring_modification(grid, some_arg)
        A dummy function to demonstrate automated docstring changes.
    <BLANKLINE>
        Construction::
    <BLANKLINE>
            grid.dummy_func_to_demonstrate_docstring_modification(some_arg)
    <BLANKLINE>
        Parameters
        ----------
        some_arg : whatever
            A dummy argument.
    <BLANKLINE>
        Examples
        --------
        ...
    <BLANKLINE>
    """
    import re
    for name, func in funcs.items():
        # strip the entry under "Parameters":
        func.__doc__ = re.sub('grid *:.*?\n.*?\n *', '', func.__doc__)
        # # cosmetic magic to get a two-line signature to line up right:
        match_2_lines = re.search(func.__name__+'\(grid,[^\)]*?\n.*?\)',
                                  func.__doc__)
        try:
            lines_were = match_2_lines.group()
        except AttributeError:  # no successful match
            pass
        else:
            end_chars = re.search('    .*?\)', lines_were).group()[4:]
            lines_are_now = re.sub('    .*?\)', '         '+end_chars,
                                   lines_were)
            func.__doc__ = (func.__doc__[:match_2_lines.start()] +
                            lines_are_now +
                            func.__doc__[match_2_lines.end():])
        # Move "grid" in signature from an arg to the class position
        func.__doc__ = re.sub(func.__name__+'\(grid, ',
                              'grid.'+func.__name__+'(',
                              func.__doc__)


def argsort_points_by_x_then_y(pts):
    """Sort points by coordinates, first x then y, returning sorted indices.

    Parameters
    ----------
    pts : Nx2 NumPy array of float
        (x,y) points to be sorted

    Returns
    -------
    indices : Nx1 NumPy array of int
        indices of sorted (x,y) points
    
    Examples
    --------
    >>> import numpy as np
    >>> from landlab.core.utils import argsort_points_by_x_then_y
    >>> pts = np.zeros((10, 2))
    >>> pts[:,0] = np.array([0., 0., 0., 1., 1., 1., 1., 2., 2., 2.])
    >>> pts[:,1] = np.array([0., 1., 2., -0.5, 0.5, 1.5, 2.5, 0., 1., 2.])
    >>> idx = argsort_points_by_x_then_y(pts)
    >>> idx
    array([3, 0, 7, 4, 1, 8, 5, 2, 9, 6])
    """
    a = pts[:,0].argsort(kind='mergesort')
    b = pts[a,1].argsort(kind='mergesort')
    return as_id_array(a[b])


def sort_points_by_x_then_y(pts):
    """Sort points by coordinates, first x then y.

    Parameters
    ----------
    pts : Nx2 NumPy array of float
        (x,y) points to be sorted

    Returns
    -------
    pts : Nx2 NumPy array of float
        sorted (x,y) points

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.core.utils import sort_points_by_x_then_y
    >>> pts = np.zeros((10, 2))
    >>> pts[:,0] = np.array([0., 0., 0., 1., 1., 1., 1., 2., 2., 2.])
    >>> pts[:,1] = np.array([0., 1., 2., -0.5, 0.5, 1.5, 2.5, 0., 1., 2.])
    >>> pts = sort_points_by_x_then_y(pts)
    >>> pts
    array([[ 1. , -0.5],
           [ 0. ,  0. ],
           [ 2. ,  0. ],
           [ 1. ,  0.5],
           [ 0. ,  1. ],
           [ 2. ,  1. ],
           [ 1. ,  1.5],
           [ 0. ,  2. ],
           [ 2. ,  2. ],
           [ 1. ,  2.5]])
    """
    indices = argsort_points_by_x_then_y(pts)
    pts[:, 0] = pts[indices, 0]
    pts[:, 1] = pts[indices, 1]
    return pts


def anticlockwise_argsort_points(pts, midpt=None):
    """Argort points into anticlockwise order around a supplied center.

    Sorts CCW from east. Assumes a convex hull.

    Parameters
    ----------
    pts : Nx2 NumPy array of float
        (x,y) points to be sorted
    midpt : len-2 NumPy array of float (optional)
        (x, y) of point about which to sort. If not provided, mean of pts is
        used.

    Returns
    -------
    pts : Nx2 NumPy array of float
        sorted (x,y) points

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.core.utils import anticlockwise_argsort_points
    >>> pts = np.zeros((4, 2))
    >>> pts[:,0] = np.array([-3., -1., -1., -3.])
    >>> pts[:,1] = np.array([-1., -3., -1., -3.])
    >>> sortorder = anticlockwise_argsort_points(pts)
    >>> np.all(sortorder == np.array([2, 0, 3, 1]))
    True
    """
    if midpt is None:
        midpt = pts.mean(axis=0)
    assert len(midpt) == 2
    theta = np.arctan2(pts[:, 1] - midpt[1], pts[:, 0] - midpt[0])
    theta = theta % (2.*np.pi)
    sortorder = np.argsort(theta)
    return sortorder


if __name__ == '__main__':
    import doctest
    doctest.testmod()



import numpy as np
from numpy.testing import assert_array_equal
from nose.tools import (assert_equal, assert_raises, raises, assert_true,
                        assert_false)
try:
    from nose.tools import assert_is
except ImportError:
    from landlab.testing.tools import assert_is

from landlab import RasterModelGrid


class TestRasterModelGridConnectingFaces():

    def setup(self):
        self.rmg = RasterModelGrid(4, 5)

    def test_horizontally_adjacent_cells(self):
        assert_array_equal(self.rmg.face_connecting_cell_pair(0, 1),
                           np.array([4]))

    def test_vertically_adjacent_cells(self):
        assert_array_equal(self.rmg.face_connecting_cell_pair(0, 3),
                           np.array([7]))

    def test_diagonally_adjacent_cells(self):
        assert_array_equal(self.rmg.face_connecting_cell_pair(1, 5),
                           np.array([]))

    def test_non_adjacent_cells(self):
        assert_array_equal(self.rmg.face_connecting_cell_pair(0, 2),
                           np.array([]))


class TestRasterModelGridCellFaces():

    def setup(self):
        self.rmg = RasterModelGrid(4, 5)

    def test_id_as_int(self):
        assert_array_equal(self.rmg.faces_at_cell[0], np.array([4, 7, 3, 0]))

    def test_id_as_array(self):
        assert_array_equal(self.rmg.faces_at_cell[[0, 1]],
                           np.array([[4, 7, 3, 0], [5, 8, 4, 1]]))
